use std::convert::TryFrom;

use chrono::NaiveDate;
use diesel::{ExpressionMethods, PgTextExpressionMethods, QueryDsl, sql_query};
use diesel::pg::types::sql_types::Array;
use diesel::sql_types::BigInt;
use rocket::State;

use crate::database::controller::connector::establish_connection;
use crate::database::controller::groups::get_groups_for_member;
use crate::database::controller::members::TempQuery;
use crate::database::model::api_members::RawMemberSearchResult;
use crate::diesel::RunQueryDsl;
use crate::helper::check_access::check_access_to_member_or_group;
use crate::helper::order_enum::Order;
use crate::helper::permission::PermissionMatrix;
use crate::helper::settings::Settings;
use crate::modules::api::members::get_member::MemberSearchResult;
use crate::modules::member_management::model::groups::Group;
use crate::schema::members;

pub fn get_raw_member_search_result(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
) -> Result<RawMemberSearchResult, diesel::result::Error> {
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let short_member: Result<RawMemberSearchResult, diesel::result::Error> = members
        .filter(entity_id.eq(member_entity_id))
        .select((entity_id, firstname, lastname))
        .first(&connection);

    match short_member {
        Ok(raw_member) => Ok(raw_member),
        Err(e) => {
            error!("Couldn't load RawMemberSearchResult: {}", e);
            return Err(e);
        }
    }
}

pub fn get_member_search_result(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
    caller_entity_id: uuid::Uuid,
) -> Result<MemberSearchResult, diesel::result::Error> {
    let member = get_raw_member_search_result(settings, member_entity_id)?;
    let groups = get_groups_for_member(settings, member_entity_id)?;

    let readable = check_access_to_member_or_group(
        settings,
        member_entity_id,
        groups.clone().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect(),
        caller_entity_id,
        crate::permissions::modules::member_management::profile::VIEW.to_string(),
    );
    Ok(MemberSearchResult {
        entity_id: member_entity_id,
        firstname: member.firstname,
        lastname: member.lastname,
        readable,
        groups,
    })
}

pub fn get_raw_member_search_result_by_name(
    settings: &State<Settings>,
    member_name: String,
) -> Result<Vec<RawMemberSearchResult>, diesel::result::Error> {
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let splitted_name: Vec<&str> = member_name.split_whitespace().collect();

    let short_members: Result<Vec<RawMemberSearchResult>, diesel::result::Error>;

    if splitted_name.len() == 0 {
        return Ok(vec![]);
    } else if splitted_name.len() == 1 {
        short_members = members
            .filter(firstname.ilike(format!("{}%", member_name)))
            .or_filter(lastname.ilike(format!("{}%", member_name)))
            .select((entity_id, firstname, lastname))
            .limit(5)
            .load(&connection);
    } else {
        short_members = members
            .filter(firstname.ilike(format!("{}%", splitted_name[0])))
            .filter(lastname.ilike(format!("{}%", splitted_name[1])))
            .select((entity_id, firstname, lastname))
            .limit(5)
            .load(&connection);
    }

    match short_members {
        Ok(raw_member) => Ok(raw_member),
        Err(e) => {
            warn!("Couldn't load RawMemberSearchResult: {}", e);
            return Err(e);
        }
    }
}

pub fn get_member_search_result_by_name(
    settings: &State<Settings>,
    name: String,
    caller_entity_id: uuid::Uuid,
) -> Result<Vec<MemberSearchResult>, diesel::result::Error> {
    let raw_members = get_raw_member_search_result_by_name(settings, name)?;
    let mut members: Vec<MemberSearchResult> = vec![];

    for member in raw_members {
        let groups = get_groups_for_member(settings, member.entity_id)?;
        let readable = check_access_to_member_or_group(
            settings,
            member.entity_id,
            groups.clone().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect(),
            caller_entity_id,
            "modules.member_management.profile.view".to_string(),
        );
        members.push(MemberSearchResult {
            entity_id: member.entity_id,
            firstname: member.firstname,
            lastname: member.lastname,
            readable,
            groups,
        });
    }

    Ok(members)
}

#[derive(Queryable, Clone, Deserialize, Serialize, QueryableByName)]
#[table_name = "members"]
pub struct MemberListEntry{
    pub(crate) entity_id: uuid::Uuid,
    pub(crate) firstname: String,
    pub(crate) lastname: String,
    pub(crate) personnel_number: Option<i32>,
    pub(crate) date_of_birth: Option<NaiveDate>,
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct MemberListEntryWithPermissions{
    pub(crate) entity_id: uuid::Uuid,
    pub(crate) firstname: String,
    pub(crate) lastname: String,
    pub(crate) personnel_number: Option<i32>,
    pub(crate) date_of_birth: Option<NaiveDate>,
    pub(crate) permissions: PermissionMatrix,
}

pub enum MemberListOrder{
    EntityID(Order),
    Firstname(Order),
    Lastname(Order),
    PersonnelNumber(Order),
    DateOfBirth(Order),
}

/// tries to parse string (syntax sort_field:[asc|desc]) into MemberListOrder
impl TryFrom<String> for MemberListOrder{
    type Error = String;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        let splits : Vec<&str> = value.split(":").collect();

        if splits.len() != 2{
            return Err(String::from("sort string invalid: no colon found."))
        }else{
            let order : Order = match splits[1].to_lowercase().as_str(){
                "asc" => Order::ASC,
                "desc" => Order::DESC,
                _ => {
                    return Err(format!("unknown order by operator {}", splits[1]))
                }
            };

            match splits[0].to_lowercase().as_str(){
                "entity_id" => Ok(MemberListOrder::EntityID(order)),
                "firstname" => Ok(MemberListOrder::Firstname(order)),
                "lastname" => Ok(MemberListOrder::Lastname(order)),
                "date_of_birth" => Ok(MemberListOrder::DateOfBirth(order)),
                "personnel_number" => Ok(MemberListOrder::PersonnelNumber(order)),
                _ => {
                    Err(format!("unknown sort field {}", splits[0]))
                }
            }
        }
    }
}

pub fn get_member_list(settings: &State<Settings>, caller: uuid::Uuid, limit: i64, offset: i64, groups: Option<Vec<uuid::Uuid>>, order_by: Option<MemberListOrder>) -> Result<Vec<MemberListEntryWithPermissions>, diesel::result::Error>{
    let connection = establish_connection(settings);

    let mut query: String = "SELECT DISTINCT members.entity_id, firstname, lastname, personnel_number, date_of_birth FROM members ".to_string();

    let mut order_query_part = String::from("");
    match order_by {
        Some(order) => {
            order_query_part += "ORDER BY ";
            match order {
                MemberListOrder::EntityID(order) => {
                    order_query_part += &format!("entity_id {} ", order.to_string())
                }
                MemberListOrder::Firstname(order) => {
                    order_query_part += &format!("firstname {} ", order.to_string());
                }
                MemberListOrder::Lastname(order) => {
                    order_query_part += &format!("lastname {} ", order.to_string());
                }
                MemberListOrder::PersonnelNumber(order) => {
                    order_query_part += &format!("personnel_number {} ", order.to_string());
                }
                MemberListOrder::DateOfBirth(order) => {
                    order_query_part += &format!("date_of_birth {} ", order.to_string());
                }
            }
        },
        None => {}
    }

    if groups.is_some(){
        query += &format!("INNER JOIN groups_entities ge on members.entity_id = ge.entity_id WHERE group_id = ANY($1) {} LIMIT $2 OFFSET $3", order_query_part);
    }else{
        query += &format!("{} LIMIT $1 OFFSET $2", order_query_part);
    }

    let res= match groups{
        Some(groups) => {
            sql_query(query).bind::<Array<diesel::pg::types::sql_types::Uuid>, _>(groups).bind::<BigInt, _>(limit).bind::<BigInt, _>(offset).get_results(&connection)
        },
        None => {
            sql_query(query).bind::<BigInt, _>(limit).bind::<BigInt, _>(offset).get_results(&connection)
        }
    };

    match res{
        Ok(memberl) => return Ok(check_permissions_for_member_list(settings, memberl, caller)),
        Err(e) => {
            warn!("Couldn't get member list: {}", e);
            return Err(e)
        }
    }
}

pub fn check_permissions_for_member_list(settings: &State<Settings>, mle: Vec<MemberListEntry>, caller: uuid::Uuid) -> Vec<MemberListEntryWithPermissions>{
    let mut res : Vec<MemberListEntryWithPermissions> = vec![];

    for entry in mle{
        let groups : Vec<Group> = get_groups_for_member(settings, entry.entity_id).unwrap().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect();
        let mut matrix = PermissionMatrix{
            read: false,
            delete: false,
            create: false,
            update: false
        };
        matrix.read = check_access_to_member_or_group(settings, entry.entity_id, groups.clone(), caller, crate::permissions::modules::member_management::profile::VIEW.to_string());
        matrix.delete = check_access_to_member_or_group(settings, entry.entity_id, groups.clone(), caller, crate::permissions::modules::member_management::profile::EDIT.to_string());
        matrix.update = check_access_to_member_or_group(settings, entry.entity_id, groups, caller, crate::permissions::modules::member_management::profile::DELETE.to_string());

        res.push(MemberListEntryWithPermissions{
            entity_id: entry.entity_id,
            firstname: entry.firstname,
            lastname: entry.lastname,
            personnel_number: entry.personnel_number,
            date_of_birth: entry.date_of_birth,
            permissions: matrix,
        })
    }
    res
}

pub fn get_member_list_count(settings: &State<Settings>, groups: Option<Vec<uuid::Uuid>>) -> Result<i64, diesel::result::Error> {
    let connection = establish_connection(settings);

    let mut query: String = "SELECT DISTINCT COUNT(members.entity_id) FROM members ".to_string();

    if groups.is_some(){
        query += "INNER JOIN groups_entities ge on members.entity_id = ge.entity_id WHERE group_id = ANY($1)";
    }

    let res: Result<TempQuery, diesel::result::Error> = match groups{
        Some(groups) => {
            sql_query(query).bind::<Array<diesel::pg::types::sql_types::Uuid>, _>(groups).get_result(&connection)
        },
        None => {
            sql_query(query).get_result(&connection)
        }
    };

    match res{
        Ok(c) => return Ok(c.count),
        Err(e) => {
            warn!("Couldn't get member list count: {}", e);
            return Err(e)
        }
    }
}

#[deprecated = "use crate::database::controller::entities::remove_entity instead"]
pub fn delete_entity_deprecated(
    settings: &State<Settings>,
    entity_id_to_delete: uuid::Uuid,
) -> Result<usize, diesel::result::Error> {
    use crate::schema::entities::dsl::*;

    let connection = establish_connection(settings);

    match diesel::delete(entities.filter(entity_id.eq(entity_id_to_delete))).execute(&connection) {
        Ok(size) => Ok(size),
        Err(e) => {
            warn!("Couldn't delete entity: {}", e);
            Err(e)
        }
    }
}
