use chrono::Utc;
use diesel::{ExpressionMethods, QueryDsl, RunQueryDsl};
use rocket::State;

use crate::database::controller::connector::establish_connection;
use crate::schema::eu_instances;
use crate::Settings;

#[derive(Queryable, Clone, Deserialize, Serialize, AsChangeset, Insertable)]
#[table_name = "eu_instances"]
#[changeset_options(treat_none_as_null = "true")]
#[primary_key(entity_id)]
pub struct RawEventUnitInstance {
    pub(crate) instance_id: uuid::Uuid,
    pub(crate) template_id: uuid::Uuid,
    pub(crate) name: String,
    pub(crate) event_id: uuid::Uuid,
    pub(crate) planned_start_time: Option<chrono::DateTime<Utc>>,
    pub(crate) planned_end_time: Option<chrono::DateTime<Utc>>,
    pub(crate) real_start_time: Option<chrono::DateTime<Utc>>,
    pub(crate) real_end_time: Option<chrono::DateTime<Utc>>,
    pub(crate) billing_rate_id: Option<uuid::Uuid>,
    pub(crate) billing_state_id: Option<uuid::Uuid>,
    pub(crate) billing_state_author: Option<uuid::Uuid>,
}

pub fn get_instances(settings: &State<Settings>, event: uuid::Uuid) -> Result<Vec<RawEventUnitInstance>, diesel::result::Error> {
    use crate::schema::eu_instances::dsl::*;

    let connection = establish_connection(settings);
    match eu_instances.filter(event_id.eq(event)).get_results(&connection) {
        Ok(instances) => Ok(instances),
        Err(e) => {
            error!("Couldn't get event unit instances: {}", e);
            Err(e)
        }
    }
}

pub fn get_instance(settings: &State<Settings>, instance: uuid::Uuid) -> Result<RawEventUnitInstance, diesel::result::Error> {
    use crate::schema::eu_instances::dsl::*;

    let connection = establish_connection(settings);
    match eu_instances.filter(instance_id.eq(instance)).get_result(&connection) {
        Ok(instances) => Ok(instances),
        Err(e) => {
            error!("Couldn't get event unit instance: {}", e);
            Err(e)
        }
    }
}


#[derive(Queryable, Clone, Deserialize, Serialize, AsChangeset, Insertable)]
#[table_name = "eu_instances"]
#[primary_key(entity_id)]
pub struct RawEventUnitInstanceChangeset {
    pub(crate) instance_id: uuid::Uuid,
    pub(crate) event_id: uuid::Uuid,
    pub(crate) name: Option<String>,
    pub(crate) planned_start_time: Option<Option<chrono::DateTime<Utc>>>,
    pub(crate) planned_end_time: Option<Option<chrono::DateTime<Utc>>>,
    pub(crate) real_start_time: Option<Option<chrono::DateTime<Utc>>>,
    pub(crate) real_end_time: Option<Option<chrono::DateTime<Utc>>>,
    pub(crate) billing_rate_id: Option<Option<uuid::Uuid>>,
    pub(crate) billing_state_id: Option<Option<uuid::Uuid>>,
}

pub fn update_instance(settings: &State<Settings>, patch: RawEventUnitInstanceChangeset) -> Result<RawEventUnitInstance, diesel::result::Error> {
    use crate::schema::eu_instances::dsl::*;

    let connection = establish_connection(settings);
    match diesel::update(eu_instances.filter(event_id.eq(patch.event_id)).filter(instance_id.eq(patch.instance_id))).set(patch).get_result(&connection) {
        Ok(update) => Ok(update),
        Err(e) => {
            error!("Couldn't patch event unit instance: {}", e);
            Err(e)
        }
    }
}

/// Replaces null values for real_start_time and real_end_time in eu_instances to the planned_start_time/planned_end_time
pub fn set_missing_real_times_to_planned_times(settings: &State<Settings>, event: uuid::Uuid) -> Result<(), diesel::result::Error> {
    use diesel::sql_query;

    let connection = establish_connection(settings);
    match sql_query("UPDATE eu_instances SET real_start_time = planned_start_time WHERE instance_id IN (SELECT instance_id FROM eu_instances WHERE event_id = $1 AND real_start_time IS NULL);").bind::<diesel::sql_types::Uuid, _>(event).execute(&connection) {
        Ok(res) => (),
        Err(e) => {
            error!("Couldn't set missing real times to planned times for instances: {}", e);
            return Err(e)
        }
    }
    match sql_query("UPDATE eu_instances SET real_end_time = planned_end_time WHERE instance_id IN (SELECT instance_id FROM eu_instances WHERE event_id = $1 AND real_end_time IS NULL);").bind::<diesel::sql_types::Uuid, _>(event).execute(&connection) {
        Ok(res) => Ok(()),
        Err(e) => {
            error!("Couldn't set missing real times to planned times for instances: {}", e);
            return Err(e)
        }
    }
}