use crate::database::controller::connector::establish_connection;
use crate::database::model::member_qualifications::RawQualificationCategory;
use crate::helper::settings::Settings;
use crate::modules::member_management::model::qualifications::{Qualification, QualificationCategory};
use crate::schema::qualification_categories::dsl::qualification_categories;
use crate::schema::qualifications_members::columns::{member_id, qualification_id};
use crate::schema::qualifications_members::dsl::qualifications_members;
use diesel::result::Error;
use diesel::{ExpressionMethods, JoinOnDsl, QueryDsl, RunQueryDsl};
use rocket::State;

pub fn get_raw_qualification_categories(
    settings: &State<Settings>,
) -> Vec<RawQualificationCategory> {
    use crate::schema::qualification_categories::dsl::*;

    let connection = establish_connection(settings);

    let data: Result<Vec<RawQualificationCategory>, diesel::result::Error> =
        qualification_categories.load(&connection);

    match data {
        Ok(data) => data,
        Err(e) => {
            error!("Couldn't read Qualification categories: {}", e);
            vec![]
        }
    }
}

pub fn get_qualifcation_categories(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
) -> Vec<QualificationCategory> {
    let raw_categories = get_raw_qualification_categories(settings);
    let mut categories: Vec<QualificationCategory> = vec![];

    for category in raw_categories {
        let qualifications =
            get_qualifications_for_member_for_category(settings, member_entity_id, category.id);
        categories.push(QualificationCategory {
            id: category.id,
            name: category.name,
            description: category.description,
            qualifications,
            visible: true,
        })
    }

    for mut category in categories.iter_mut() {
        if category.qualifications.is_empty() {
            category.visible = false;
        }
    }

    categories
}

pub fn get_qualifications_for_member_for_category(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
    category_id: uuid::Uuid,
) -> Vec<Qualification> {
    use crate::schema::qualifications::dsl::*;
    use crate::schema::qualifications_members::dsl::*;

    let connection = establish_connection(settings);

    let data: Result<Vec<Qualification>, diesel::result::Error> = qualifications
        .left_join(qualifications_members.on(qualification_id.eq(id)))
        .filter(member_id.eq(member_entity_id))
        .filter(category.eq(category_id))
        .select((id, name, description))
        .load(&connection);

    match data {
        Ok(data) => data,
        Err(e) => {
            error!(
                "Couldn't read Qualifications for category {} for member {}: {}",
                category_id, member_entity_id, e
            );
            vec![]
        }
    }
}

pub fn remove_qualification_for_member(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
    qualification_id2: uuid::Uuid,
) -> Result<usize, Error> {
    let connection = establish_connection(settings);
    let result = diesel::delete(
        qualifications_members
            .filter(member_id.eq(member_entity_id))
            .filter(qualification_id.eq(qualification_id2)),
    )
    .execute(&connection);

    match result {
        Ok(result) => Ok(result),
        Err(e) => {
            error!("Couldn't delete Qualifications {}", e);
            Err(e)
        }
    }
}

pub fn add_qualification_for_member(
    settings: &State<Settings>,
    member_entity_id: uuid::Uuid,
    qualification_id2: uuid::Uuid,
) -> Result<usize, Error> {
    let connection = establish_connection(settings);
    let result = diesel::insert_into(qualifications_members)
        .values((
            member_id.eq(member_entity_id),
            qualification_id.eq(qualification_id2),
        ))
        .execute(&connection);

    match result {
        Ok(result) => Ok(result),
        Err(e) => {
            error!("Couldn't insert Qualification {}", e);
            Err(e)
        }
    }
}

pub fn get_qualifications_for_category(
    settings: &State<Settings>,
    category_id2: uuid::Uuid,
) -> Vec<Qualification> {
    use crate::schema::qualifications::dsl::*;

    let connection = establish_connection(settings);

    let data: Result<Vec<Qualification>, diesel::result::Error> = qualifications
        .left_join(qualification_categories.on(category.eq(id)))
        .filter(category.eq(category_id2))
        .select((id, name, description))
        .load(&connection);

    match data {
        Ok(data) => data,
        Err(e) => {
            error!(
                "Couldn't read Qualifications for category {} : {}",
                category_id2, e
            );
            vec![]
        }
    }
}

pub fn get_qualifications(settings: &State<Settings>) -> Result<Vec<QualificationCategory>, diesel::result::Error>{
    use crate::schema::qualification_categories::dsl::*;
    use crate::schema::qualifications::dsl::*;

    let connection = establish_connection(settings);
    let mut categories_with_qualifications : Vec<QualificationCategory> = vec![];

    let categories : Vec<RawQualificationCategory> = match qualification_categories.load(&connection){
        Ok(cat) => cat,
        Err(e) => {
            error!("Couldn't get qualification categories: {}", e);
            return Err(e)}
    };
    for cat in categories{
        let res : Vec<Qualification> = match qualifications.filter(category.eq(cat.id)).select((crate::schema::qualifications::dsl::id, crate::schema::qualifications::dsl::name, crate::schema::qualifications::dsl::description)).get_results(&connection){
            Ok(res) => res,
            Err(e) => {
                error!("Couldn't get qualifications: {}", e);
                return Err(e)
            }
        };
        categories_with_qualifications.push(QualificationCategory{
            id: cat.id,
            name: cat.name,
            description: cat.description,
            qualifications: res,
            visible: false
        })
    }

    Ok(categories_with_qualifications)
}

pub fn check_qualification_for_member(settings: &State<Settings>, qualification: uuid::Uuid, member: uuid::Uuid) -> Result<bool, diesel::result::Error>{
    use diesel::dsl::count;

    let connection = establish_connection(settings);

    match qualifications_members.filter(member_id.eq(member)).filter(qualification_id.eq(qualification)).select(count(member_id)).get_result::<i64>(&connection){
        Ok(res) => if res >= 1 {
            Ok(true)
        }else{
            Ok(false)
        },
        Err(e) => {
            error!("Couldn't check license for member: {}", e);
            Err(e)
        }
    }
}