use diesel::{JoinOnDsl, QueryDsl, sql_query};
use diesel::pg::expression::array_comparison::any;
use diesel::pg::types::sql_types::{Array, Uuid};
use diesel::sql_types::{BigInt, Text};
use rocket::State;

use crate::database::controller::connector::establish_connection;
use crate::database::controller::groups::get_groups_for_member;
use crate::database::model::members::RawMember;
use crate::diesel::prelude::*;
use crate::helper::settings::Settings;
use crate::modules::member_management::model::member::Member;
use crate::schema::groups_entities::dsl::groups_entities;

/*impl<DB> FromSql<SmallInt, DB> for Sex
where
    DB: Backend,
    i16: FromSql<SmallInt, DB>,
{
    fn from_sql(bytes: Option<&DB::RawValue>) -> deserialize::Result<Self> {
        match i16::from_sql(bytes)? {
            0 => Ok(Sex::UNKNOWN),
            1 => Ok(Sex::MALE),
            2 => Ok(Sex::FEMALE),
            9 => Ok(Sex::OTHER),
            x => Err(format!("Unrecognized variant {}", x).into()),
        }
    }
}*/

/// Aggregates all permissions a member has (from all assigned roles)
///
/// #Arguments
/// * 'uuid' - Member's uuid, type uuid::Uuid
/// * 'settings' - Settings, as managed State
pub fn get_member_permissions_by_uuid(
    uuid: uuid::Uuid,
    settings: &State<Settings>,
) -> Option<Vec<String>> {
    use crate::schema::members::dsl::*;
    use crate::schema::members_roles::dsl::*;
    use crate::schema::roles_permissions::dsl::permission_id;
    use crate::schema::roles_permissions::dsl::role_id as role_id2;
    use crate::schema::roles_permissions::dsl::roles_permissions;

    let connection = establish_connection(settings);

    let data = members
        .left_join(members_roles.on(member_id.eq(entity_id)))
        .left_join(roles_permissions.on(role_id2.eq(role_id)))
        .select(permission_id.nullable())
        .filter(entity_id.eq(uuid))
        .load::<Option<String>>(&connection);

    let data = match data {
        Ok(data) => data,
        Err(e) => {
            error!(
                "Error occured while checking permissions for member {}: {}",
                uuid, e
            );
            return None;
        }
    };

    let mut permissions: Vec<String> = vec![];

    for permission in data {
        match permission {
            Some(permission) => permissions.push(permission),
            None => (),
        }
    }

    if !permissions.is_empty() {
        Some(permissions)
    } else {
        None
    }
}

/// Get vector of all members (database::model::members::RawMember) for a user uuid
///
/// #Arguments
/// * 'uuid' - User's uuid, type uuid::Uuid
/// * 'settings' - Settings, as managed State
///
/// #Returns
/// * 'Vec<RawMember>
pub fn get_raw_members_by_user_uuid(
    uuid: uuid::Uuid,
    settings: &State<Settings>,
) -> Vec<RawMember> {
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let member: Result<Vec<RawMember>, diesel::result::Error> =
        members.filter(users_id.eq(uuid)).load(&connection);
    match member {
        Ok(member) => member,
        Err(e) => {
            error!(
                "Error occured while loading data for member {}: {}",
                uuid, e
            );
            vec![]
        }
    }
}

/// Get RawMember (database::model::members::RawMember) for a member uuid
///
/// #Arguments
/// * 'uuid' - Member's uuid, type uuid::Uuid
/// * 'settings' - Settings, as managed State
///
/// #Returns
/// * 'Option<RawMember>'
pub fn get_raw_member_by_uuid(uuid: uuid::Uuid, settings: &State<Settings>) -> Option<RawMember> {
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let member: Result<RawMember, diesel::result::Error> =
        members.filter(entity_id.eq(uuid)).first(&connection);
    match member {
        Ok(member) => Some(member),
        Err(e) => {
            error!(
                "Error occured while loading data for member {}: {}",
                uuid, e
            );
            None
        }
    }
}

/// Get vector of all members (database::model::members::RawMember)
///
/// #Arguments
/// * 'settings' - Settings, as managed State
///
/// #Returns
/// * 'Vec<RawMember>'
pub fn get_raw_members(settings: &State<Settings>) -> Vec<RawMember> {
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let member: Result<Vec<RawMember>, diesel::result::Error> = members.load(&connection);
    match member {
        Ok(member) => member,
        Err(e) => {
            error!("Error occured while loading data for member: {}", e);
            vec![]
        }
    }
}

/// Get vector of all members (database::model::members::RawMember)
///
/// #Arguments
/// * 'settings' - Settings, as managed State
/// * 'group_list' - list of all groups selected as Vec<uuid::Uuid>
///
/// #Returns
/// * 'Vec<RawMember>'
pub fn get_raw_members_in_group(
    settings: &State<Settings>,
    group_list: Vec<uuid::Uuid>,
) -> Vec<RawMember> {
    use crate::schema::groups_entities::dsl::entity_id as group_entity_id;
    use crate::schema::groups_entities::dsl::*;
    use crate::schema::members::dsl::entity_id as member_entity_id;
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let data: Result<Vec<RawMember>, diesel::result::Error> = members
        .left_join(groups_entities.on(member_entity_id.eq(group_entity_id)))
        .filter(group_id.eq(any(group_list)))
        .select((
            member_entity_id,
            users_id,
            firstname,
            lastname,
            date_of_birth,
            sex,
            salutation,
            place_of_birth,
            academic_titles,
            personnel_number,
            ui_language,
            nationality,
            entrance_date,
            birth_name,
            iban,
            bic
        ))
        .distinct()
        .order(lastname.asc())
        .load(&connection);

    match data {
        Ok(data) => data,
        Err(e) => {
            error!("Error occured while loading data for member: {}", e);
            vec![]
        }
    }
}

/// Get member_management::model::member::Member
///
/// #Arguments
/// * 'uuid' - Users's uuid, type uuid::Uuid
/// * 'settings' - Settings, as managed State
/// #Returns
/// * 'Vec<Member>'
pub fn get_members_by_user_uuid(uuid: uuid::Uuid, settings: &State<Settings>) -> Vec<Member> {
    let raw_members: Vec<RawMember> = get_raw_members_by_user_uuid(uuid, settings);
    let mut members: Vec<Member> = vec![];

    for raw_member in raw_members {
        let permissions = get_member_permissions_by_uuid(raw_member.entity_id, settings);
        let permissions = permissions.unwrap_or(vec![]);

        let groups = get_groups_for_member(settings, raw_member.entity_id).unwrap().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect();
        members.push(Member::from((raw_member, permissions, groups)));
    }

    members
}

/// Get member_management::model::member::Member
///
/// Will return empty permissions/groups vec if error occurred while fetching permissions/groups
/// #Arguments
/// * 'uuid' - Member's uuid, type uuid::Uuid
/// * 'settings' - Settings, as managed State
/// #Returns
/// * 'Member'
//TODO: return Result not OPtion
pub fn get_member_by_uuid(uuid: uuid::Uuid, settings: &State<Settings>) -> Option<Member> {
    let raw_member: RawMember = get_raw_member_by_uuid(uuid, &settings)?;

    let permissions = get_member_permissions_by_uuid(raw_member.entity_id, settings);
    let permissions = permissions.unwrap_or(vec![]);

    let groups = get_groups_for_member(settings, raw_member.entity_id).unwrap().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect();
    Some(Member::from((raw_member, permissions, groups)))
}

/// Get all member_management::model::member::Member member has access to (any role)
///
/// #Arguments
/// * 'settings' - Settings, as managed State
/// * 'group_list' - list of all groups selected as Vec<uuid::Uuid>
///
/// #Returns
/// * 'Vec<Member'
pub fn get_members(
    settings: &State<Settings>,
    group_list: Vec<uuid::Uuid>,
    _own_member: Member,
    limit: i64,
    offset: i64
) -> Result<Vec<Member>, diesel::result::Error> {
    let raw_members: Vec<RawMember> = match get_member_list(settings, group_list, limit, offset){
        Ok(members) => members,
        Err(e) => {
            return Err(e)
        }
    };
    let mut members: Vec<Member> = vec![];

    for raw_member in raw_members {
        let permissions = get_member_permissions_by_uuid(raw_member.entity_id, settings);
        let permissions = permissions.unwrap_or(vec![]);
        let groups = get_groups_for_member(settings, raw_member.entity_id).unwrap().into_iter().map(|m|crate::modules::member_management::model::groups::Group::from(m)).collect();
        let member = Member::from((raw_member, permissions, groups));

        members.push(member);
    }

    Ok(members)
}

#[derive(QueryableByName, PartialEq, Debug)]
pub(crate) struct TempQuery {
    #[sql_type = "BigInt"]
    pub(crate) count: i64,
}

/// Checks if member has permission on specific resource
///
/// * 'settings' - Settings, as managed State
/// * 'caller_id' - UUID of caller requesting access
/// * 'entity' - UUID of entity caller requesting access to
/// * 'permission' - permission caller tries to use on entity
pub fn check_access_to_resource(
    settings: &State<Settings>,
    caller_id: uuid::Uuid,
    entity: uuid::Uuid,
    permission: &str,
) -> bool {
    let connection = establish_connection(settings);

    let permission_count: Result<TempQuery, diesel::result::Error> = sql_query(
        "SELECT COUNT(role_permission_id) FROM roles_permissions_context WHERE entity = $1 AND role_permission_id IN (SELECT role_permission_id FROM roles_permissions WHERE role_id IN (SELECT role_id FROM members_roles WHERE member_id = $2) AND permission_id = $3)",
    )
    .bind::<Uuid, _>(entity)
    .bind::<Uuid, _>(caller_id)
    .bind::<Text, _>(permission)
    .get_result(&connection);

    match permission_count {
        Ok(count) => {
            if count.count > 0 {
                true
            } else {
                false
            }
        }
        Err(e) => {
            error!("Error while checking members permissions (context): {}", e);
            false
        }
    }
}

pub fn get_member_list(settings: &State<Settings>, selected_groups: Vec<uuid::Uuid>, limit: i64, offset: i64) -> Result<Vec<RawMember>, diesel::result::Error>{
    use crate::schema::members::dsl::*;
    let connection = establish_connection(settings);

    let memberlist : Result<Vec<RawMember>, diesel::result::Error> = crate::schema::members::dsl::members.left_join(groups_entities.on(crate::schema::members::entity_id.eq(crate::schema::groups_entities::entity_id))).filter(crate::schema::groups_entities::group_id.eq(any(selected_groups))).select((entity_id, users_id, firstname, lastname, date_of_birth, sex, salutation, place_of_birth, academic_titles, personnel_number, ui_language, nationality, entrance_date, birth_name, iban, bic)).order(lastname.asc()).distinct().limit(limit).offset(offset).load(&connection);
    match memberlist{
        Ok(memberlist) => Ok(memberlist),
        Err(e) => {
            error!("Couldn't get member list: {}", e);
            Err(e)
        }
    }
}

pub fn member_count(settings: &State<Settings>, selected_groups: Vec<uuid::Uuid>) -> Result<i64, diesel::result::Error>{
    let connection = establish_connection(settings);

    let count : Result<TempQuery, diesel::result::Error> = sql_query("SELECT count(entity_id) FROM members WHERE entity_id IN (SELECT entity_id FROM groups_entities WHERE group_id = ANY($1))").bind::<Array<Uuid>, _>(selected_groups).get_result(&connection);
    match count{
        Ok(c) => Ok(c.count),
        Err(e) => {
            error!("Couldn't count members: {}", e);
            Err(e)
        }
    }
}