use crate::database::controller::api_members::get_member_search_result;
use crate::database::controller::connector::establish_connection;
use crate::database::controller::groups::GroupEntityState;
use crate::database::model::api_members::RawMemberSearchResult;
use crate::helper::settings::Settings;
use crate::modules::api::members::get_member::MemberSearchResult;
use diesel::{ExpressionMethods, JoinOnDsl, QueryDsl, RunQueryDsl};
use diesel::result::DatabaseErrorKind;
use rocket::State;

pub fn get_raw_member_search_results_in_group(
    settings: &State<Settings>,
    group_entity_id: uuid::Uuid,
) -> Result<Vec<RawMemberSearchResult>, diesel::result::Error> {
    use crate::schema::groups_entities::dsl::*;
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    let memberlist: Result<Vec<RawMemberSearchResult>, diesel::result::Error> =
        members
            .left_join(groups_entities.on(
                crate::schema::groups_entities::entity_id.eq(crate::schema::members::entity_id),
            ))
            .filter(group_id.eq(group_entity_id))
            .select((crate::schema::members::entity_id, firstname, lastname))
            .load(&connection);

    match memberlist {
        Ok(list) => Ok(list),
        Err(e) => {
            error!("Couldn't get members for group: {}", e);
            Err(e)
        }
    }
}

#[derive(Serialize, Deserialize, Queryable, Clone)]
pub struct MemberGroupView {
    pub(crate) entity_id: uuid::Uuid,
    pub(crate) group_id: uuid::Uuid,
    pub(crate) group_state: Option<uuid::Uuid>,
    pub(crate) firstname: String,
    pub(crate) lastname: String,
    pub(crate) readable: bool,
}

fn convert_to_group_view(msr: MemberSearchResult, group: uuid::Uuid) -> MemberGroupView {
    for single_group in msr.groups {
        if single_group.group_id == group {
            return MemberGroupView {
                entity_id: msr.entity_id,
                group_id: group,
                group_state: single_group.state,
                firstname: msr.firstname,
                lastname: msr.lastname,
                readable: msr.readable,
            };
        }
    }
    MemberGroupView {
        entity_id: msr.entity_id,
        group_id: group,
        group_state: None,
        firstname: msr.firstname,
        lastname: msr.lastname,
        readable: msr.readable,
    }
}

pub fn get_members_in_group(
    settings: &State<Settings>,
    group_entity_id: uuid::Uuid,
    caller_entity_id: uuid::Uuid,
) -> Result<Vec<MemberGroupView>, diesel::result::Error> {
    let raw_memberlist = get_raw_member_search_results_in_group(settings, group_entity_id)?;
    let mut memberlist: Vec<MemberGroupView> = vec![];

    for member in raw_memberlist {
        memberlist.push(convert_to_group_view(get_member_search_result(
            settings,
            member.entity_id,
            caller_entity_id,
        )?, group_entity_id))
    }

    Ok(memberlist)
}

pub fn add_member_to_group(
    settings: &State<Settings>,
    member_id: uuid::Uuid,
    group_id2: uuid::Uuid,
) -> Result<(), diesel::result::Error> {
    use crate::schema::groups_entities::dsl::*;

    let connection = establish_connection(settings);

    match diesel::insert_into(groups_entities)
        .values((group_id.eq(&group_id2), entity_id.eq(&member_id)))
        .execute(&connection)
    {
        Ok(_) => Ok(()),
        Err(e) => {
            error!(
                "Couldn't add member {} to group {}: {}",
                member_id, group_id2, e
            );
            Err(e)
        }
    }
}

/// Retrieves specified state from database
/// Parameters:
/// * state: state_id as [uuid::Uuid]
pub fn get_state(settings: &State<Settings>, state: uuid::Uuid) -> Result<GroupEntityState, diesel::result::Error>{
    use crate::schema::group_entity_state::dsl::*;

    let connection = establish_connection(settings);
    match group_entity_state.filter(state_id.eq(state)).get_result(&connection){
        Ok(state) => Ok(state),
        Err(e) => {
            warn!("Couldn't retrieve state: {}", e);
            Err(e)
        }
    }
}

use diesel::dsl::count;

/// Checks if state is unique for member
/// Returns Ok(true) if state is unique for member, Ok(false) if state isn't unique.
pub fn check_state_unique(settings: &State<Settings>, member: uuid::Uuid, group: uuid::Uuid, state_id: uuid::Uuid) -> Result<bool, diesel::result::Error>{
    use crate::schema::groups_entities::dsl::*;

    let connection = establish_connection(settings);
    match groups_entities.filter(state.eq(state_id)).filter(group_id.ne(group)).filter(entity_id.eq(member)).select(count(state)).get_result::<i64>(&connection){
        Ok(count) => {
            if count == 0 {
                Ok(true)
            }else{
                Ok(false)
            }
        },
        Err(e) => {
            error!("Couldn't check if state is unique for memer: {}", e);
            Err(e)
        }
    }
}

/// Update members state for group
pub fn add_member_to_state(settings: &State<Settings>, member: uuid::Uuid, group: uuid::Uuid, state_id: Option<uuid::Uuid>) -> Result<(), diesel::result::Error>{
    use crate::schema::groups_entities::dsl::*;

    let connection = establish_connection(settings);

    // Check if state has unique flag and check for uniqueness
    match state_id{
        Some(state_id) => {
            let res = get_state(settings, state_id)?;
            if res.unique{
                match check_state_unique(settings, member, group, state_id){
                    Ok(uni) => {
                        if !uni{
                            return Err(diesel::result::Error::DatabaseError(DatabaseErrorKind::UniqueViolation, Box::new(String::from("Violates uniqueness for state. There may be only one group where this entity can have this state."))))
                        }
                    }
                    Err(e) => return Err(e)
                }
            }
        }
        None => {}
    }

    match diesel::update(groups_entities.filter(entity_id.eq(member)).filter(group_id.eq(group))).set(state.eq(state_id)).execute(&connection){
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't change state for member: {}", e);
            Err(e)
        }
    }
}

pub fn remove_member_from_group(
    settings: &State<Settings>,
    member_id: uuid::Uuid,
    group_id2: uuid::Uuid,
) -> Result<(), diesel::result::Error> {
    use crate::schema::groups_entities::dsl::*;

    let connection = establish_connection(settings);

    match diesel::delete(groups_entities)
        .filter(group_id.eq(group_id2))
        .filter(entity_id.eq(member_id))
        .execute(&connection)
    {
        Ok(_) => Ok(()),
        Err(e) => {
            error!(
                "Couldn't remove member {} from group {}: {}",
                member_id, group_id2, e
            );
            Err(e)
        }
    }
}
