use rocket::State;
use crate::helper::settings::Settings;
use crate::database::model::permissions::Permission;
use crate::database::controller::connector::establish_connection;
use diesel::{ExpressionMethods, QueryDsl, RunQueryDsl, OptionalExtension, sql_query};
use diesel::sql_types::Text;
use diesel::pg::types::sql_types::Uuid;

/// Retrieves permission row for specified permission
///
/// # Returns:
// /// * Ok(Some([`crate::database::model::permissions::Permission`])) if permission found
// /// * Ok(None) if no permission found but query returned no error
// /// * Err([`diesel::result::Error`]) if query returned error
pub fn get_permission(settings: &State<Settings>, permission_str: &str) -> Result<Option<Permission>, diesel::result::Error>{
    use crate::schema::permissions::dsl::*;

    let connection = establish_connection(settings);

    match permissions.filter(permission.eq(permission_str)).get_result(&connection).optional(){
        Ok(per) => Ok(per),
        Err(e) => {
            error!("Couldn't get permission: {}", e);
            Err(e)
        }
    }

}

/// Retrieve role_permission_id for specified permission/role combo
///
/// # Returns:
/// * Ok(Some(role_permission_id)) if role_permission_id found
/// * Ok(None) if no role_permission_id found but query succeeded
/// * Err([`diesel::result::Error`]) if query returned error
///
pub fn get_role_permission_id(settings: &State<Settings>, permission_str: &str, role: &str) -> Result<Option<uuid::Uuid>, diesel::result::Error>{
    use crate::schema::roles_permissions::dsl::*;

    let connection = establish_connection(settings);

    match roles_permissions.filter(permission_id.eq(permission_str)).filter(role_id.eq(role)).select(role_permission_id).get_result(&connection).optional(){
        Ok(rpi) => Ok(rpi),
        Err(e) => {
            error!("Couldn't get role permission id: {}", e);
            Err(e)
        }
    }

}

pub fn get_role_permission_context(settings: &State<Settings>, role_permission_id: uuid::Uuid) -> Result<Vec<uuid::Uuid>, diesel::result::Error>{
    use crate::schema::roles_permissions_context::dsl::{roles_permissions_context, entity};
    use crate::schema::roles_permissions_context::dsl::role_permission_id as rpi;

    let connection = establish_connection(settings);

    match roles_permissions_context.filter( rpi.eq(role_permission_id)).select(entity).get_results(&connection){
        Ok(res) => Ok(res),
        Err(e) => {
            error!("Couldn't get list of entities for role_permission_id: {}", e);
            Err(e)
        }
    }
}

#[derive(QueryableByName, PartialEq, Debug)]
struct TempQuery {
    #[sql_type = "Uuid"]
    pub(crate) member_id: uuid::Uuid,
}

pub fn get_members_with_permission(settings: &State<Settings>, permission: &str) -> Result<Vec<uuid::Uuid>, diesel::result::Error>{
    let connection = establish_connection(settings);

    let res : Result<Vec<TempQuery>, diesel::result::Error> = sql_query("SELECT member_id FROM members_roles WHERE role_id IN (SELECT role_id FROM roles_permissions WHERE permission_id = $1);").bind::<Text, _>(permission
    ).get_results(&connection);

    match res{
        Ok(res) => Ok(res.iter().map(|tq|tq.member_id).collect()),
        Err(e) => {
            error!("Couldn't get members with permission: {}", e);
            Err(e)
        }
    }
}

pub fn get_permissions(settings: &State<Settings>) -> Result<Vec<Permission>, diesel::result::Error>{
    use crate::schema::permissions::dsl::*;

    let connection =establish_connection(settings);
    match permissions.load(&connection){
        Ok(per) => Ok(per),
        Err(e) => {
            error!("Couldn't get permission list: {}", e);
            Err(e)
        }
    }
}

pub fn add_permission_to_role(settings: &State<Settings>, permission: &str, role: &str) -> Result<uuid::Uuid, diesel::result::Error>{
    use crate::schema::roles_permissions::dsl::*;

    let connection = establish_connection(settings);
    match diesel::insert_into(roles_permissions).values((&role_id.eq(role), &permission_id.eq(permission))).returning(role_permission_id).get_result(&connection) {
        Ok(res) => Ok(res),
        Err(e) => {error!("Couldn't add permission to role: {}", e); Err(e)}
    }
}

pub fn remove_permission_from_role(settings: &State<Settings>, permission: &str, role: &str) -> Result<(), diesel::result::Error>{
    use crate::schema::roles_permissions::dsl::*;

    let connection = establish_connection(settings);
    match diesel::delete(roles_permissions).filter(permission_id.eq(permission)).filter(role_id.eq(role)).execute(&connection) {
        Ok(_) => Ok(()),
        Err(e) => {error!("Couldn't remove permission from role: {}", e); Err(e)}
    }
}