use crate::database::controller::connector::establish_connection;
use crate::database::model::roles::Role;
use crate::helper::settings::Settings;
use diesel::{BoolExpressionMethods, ExpressionMethods, RunQueryDsl, QueryDsl, JoinOnDsl};
use rocket::State;
use crate::database::model::permissions::Permission;
use crate::database::model::api_members::RawMemberSearchResult;

pub fn get_roles(settings: &State<Settings>) -> Result<Vec<Role>, diesel::result::Error> {
    use crate::schema::roles::dsl::*;

    let connection = establish_connection(settings);
    match roles.load::<Role>(&connection) {
        Ok(roles_list) => Ok(roles_list),
        Err(e) => {
            error!("Couldn't load roles: {}", e);
            Err(e)
        }
    }
}

pub fn get_roles_for_member(settings: &State<Settings>, member: uuid::Uuid) -> Result<Vec<String>, diesel::result::Error>{
    use crate::schema::members_roles::dsl::*;

    let connection = establish_connection(settings);

    match members_roles.filter(member_id.eq(member)).select(role_id).get_results(&connection){
        Ok(res) => Ok(res),
        Err(e) => {error!("Couldn't get roles for member: {}", e); Err(e)}
    }
}

pub fn get_members_for_role(settings: &State<Settings>, role: &str) -> Result<Vec<RawMemberSearchResult>, diesel::result::Error>{
    use crate::schema::members_roles::dsl::*;
    use crate::schema::members::dsl::*;

    let connection = establish_connection(settings);

    match members_roles.inner_join(members.on(member_id.eq(entity_id))).filter(role_id.eq(role)).select((crate::schema::members::dsl::entity_id, firstname, lastname)).get_results(&connection){
        Ok(res) => Ok(res),
        Err(e) => {error!("Couldn't get members in role: {}", e); Err(e)}
    }
}

/// DEPRECATED DO NOT USE, WILL BE REMOVED SOON
/// Use [crate::database::controller:permissions::get_role_permission_id] instead!
pub fn get_role_permission_id_deprecated(
    settings: &State<Settings>,
    role_id2: String,
    permission: &str,
) -> Option<uuid::Uuid> {
    use crate::schema::roles_permissions::dsl::*;

    let connection = establish_connection(settings);
    let rpi: Result<uuid::Uuid, diesel::result::Error> = roles_permissions
        .filter(role_id.eq(role_id2).and(permission_id.eq(permission)))
        .select(role_permission_id)
        .first(&connection);

    match rpi {
        Ok(rpi) => Some(rpi),
        Err(error) => match error {
            diesel::result::Error::NotFound => None,
            _ => {
                error!("Couldn't get role permission id: {}", error);
                None
            }
        },
    }
}

pub fn add_permission(
    settings: &State<Settings>,
    role_id2: String,
    permission: String,
) -> Result<uuid::Uuid, diesel::result::Error> {
    use crate::schema::roles_permissions::dsl::*;
    let connection = establish_connection(settings);

    let inserted = diesel::insert_into(roles_permissions)
        .values((role_id.eq(role_id2), permission_id.eq(permission)))
        .returning(role_permission_id)
        .get_result(&connection);

    match inserted {
        Ok(id) => Ok(id),
        Err(e) => {
            error!("Couldn't add permission: {}", e);
            Err(e)
        }
    }
}

pub fn add_permission_context(
    settings: &State<Settings>,
    role_permission_id2: uuid::Uuid,
    context: uuid::Uuid,
) -> Result<(), diesel::result::Error> {
    use crate::schema::roles_permissions_context::dsl::*;
    let connection = establish_connection(settings);

    let inserted = diesel::insert_into(roles_permissions_context)
        .values((
            role_permission_id.eq(role_permission_id2),
            entity.eq(context),
        ))
        .execute(&connection);
    match inserted {
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't add permission context: {}", e);
            Err(e)
        }
    }
}

pub fn remove_permission_context(
    settings: &State<Settings>,
    role_permission_id2: uuid::Uuid,
    context: uuid::Uuid,
) -> Result<(), diesel::result::Error> {
    use crate::schema::roles_permissions_context::dsl::*;
    let connection = establish_connection(settings);

    let deleted = diesel::delete(roles_permissions_context)
        .filter(
            role_permission_id.eq(role_permission_id2)).filter(entity.eq(context)).execute(&connection);
    match deleted {
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't delete permission context: {}", e);
            Err(e)
        }
    }
}

pub fn add_member_to_role(settings: &State<Settings>, member_uuid: uuid::Uuid, role: String) -> Result<(), diesel::result::Error>{
    use crate::schema::members_roles::dsl::*;
    let connection = establish_connection(settings);

    match diesel::insert_into(members_roles).values((member_id.eq(member_uuid), role_id.eq(role))).execute(&connection){
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't add member to role: {}", e);
            Err(e)
        }
    }
}

pub fn remove_member_from_role(settings: &State<Settings>, member_uuid: uuid::Uuid, role: String) -> Result<(), diesel::result::Error>{
    use crate::schema::members_roles::dsl::*;
    let connection = establish_connection(settings);

    match diesel::delete(members_roles).filter(member_id.eq(member_uuid)).filter(role_id.eq(role)).execute(&connection){
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't remove member from role: {}", e);
            Err(e)
        }
    }
}

pub fn get_permissions_for_role(settings: &State<Settings>, role: String) -> Result<Vec<Permission>, diesel::result::Error>{
    use crate::schema::permissions::dsl::*;

    let connection = establish_connection(settings);

    match crate::schema::roles_permissions::dsl::roles_permissions.inner_join(permissions.on(crate::schema::roles_permissions::dsl::permission_id.eq(permission))).filter(crate::schema::roles_permissions::dsl::role_id.eq(role)).select((permission, description, context, context_type)).get_results(&connection){
        Ok(res) => Ok(res),
        Err(e) => {
            error!("Couldn't get permissions for role: {}", e);
            Err(e)
        }
    }
}

pub fn update_role(settings: &State<Settings>, role_id: String, role: Role) -> Result<Role, diesel::result::Error>{
    use crate::schema::roles::dsl::*;

    let connection = establish_connection(settings);

    match diesel::update(roles).filter(id.eq(role_id)).set(role).get_result(&connection){
        Ok(res) => Ok(res),
        Err(e) => {
            error!("Couldn't update role: {}", e);
            Err(e)
        }
    }
}