use crate::helper::settings::Settings;
use rocket::State;
use crate::database::model::vehicles::{VehicleCategory, Vehicle};
use crate::database::controller::connector::establish_connection;
use diesel::{QueryDsl, RunQueryDsl, ExpressionMethods, PgTextExpressionMethods};

pub fn get_vehicle_categories(settings: &State<Settings>) -> Result<Vec<VehicleCategory>, diesel::result::Error> {
    use crate::schema::vehicle_categories::dsl::*;
    let connection = establish_connection(settings);

    match vehicle_categories.load(&connection){
        Ok(categories) => Ok(categories),
        Err(e) => {
            error!("Couldn't retrieve vehicle categories: {}", e);
            Err(e)
        }
    }
}

pub fn get_vehicles(settings: &State<Settings>, limit: i64, offset: i64, q: Option<String>) -> Result<Vec<Vehicle>, diesel::result::Error>{
    use crate::schema::vehicles::dsl::*;

    let connection = establish_connection(settings);

    let mut query = vehicles.order(identifier.asc()).limit(limit).offset(offset).into_boxed();

    match q{
        Some(q) => {
            query = query.filter(identifier.ilike(format!("%{}%", q)))
        },
        None => {}
    }

    match query.load(&connection){
        Ok(vehicle_list) => Ok(vehicle_list),
        Err(e) => {
            error!("Couldn't retrieve vehicles: {}", e);
            Err(e)
        }
    }
}

pub fn get_vehicle(settings: &State<Settings>, id: uuid::Uuid) -> Result<Vehicle, diesel::result::Error>{
    use crate::schema::vehicles::dsl::*;

    let connection = establish_connection(settings);

    match vehicles.filter(entity_id.eq(id)).get_result(&connection){
        Ok(vehicle) => Ok(vehicle),
        Err(e) => {
            error!("Couldn't retrieve vehicle: {}", e);
            Err(e)
        }
    }
}


pub fn get_vehicle_count(settings: &State<Settings>) -> Result<i64, diesel::result::Error>{
    use crate::schema::vehicles::dsl::*;
    use diesel::dsl::count;

    let connection = establish_connection(settings);

    match vehicles.select(count(entity_id)).first(&connection){
        Ok(count) => Ok(count),
        Err(e) => {
            error!("Couldn't count vehicles: {}", e);
            Err(e)
        }
    }
}

pub fn change_vehicle(settings: &State<Settings>, vehicle: Vehicle) -> Result<(), diesel::result::Error>{
    use crate::schema::vehicles::dsl::*;

    let connection = establish_connection(settings);
    match diesel::update(vehicles).filter(entity_id.eq(vehicle.entity_id)).set(&vehicle).execute(&connection){
        Ok(_) => Ok(()),
        Err(e) => {
            error!("Couldn't change vehicle: {}", e);
            Err(e)
        }
    }

}

/// Adds vehicle to database
///
/// Provide a vehicle via Vehicle struct
/// The vehicle must have an entity_id, e.g. generated with generate_entity()
pub fn add_vehicle(settings: &State<Settings>, vehicle: Vehicle) -> Result<Vehicle, diesel::result::Error>{
    use crate::schema::vehicles::dsl::*;

    let connection = establish_connection(settings);

    match diesel::insert_into(vehicles).values(&vehicle).get_result(&connection){
        Ok(vehicle) => Ok(vehicle),
        Err(e) => {
            error!("Couldn't insert new vehicle: {}", e);
            Err(e)
        }
    }
}