use std::collections::VecDeque;
use std::ops::Add;
use std::sync::RwLock;

use chrono::{Duration, Utc};
use lettre::Message;

use crate::helper::mail_queue::worker::send_mail;
use crate::Settings;

pub struct MessageWatcher {
    message: Message,
    tries: i8,
    next_try_after: Option<chrono::DateTime<Utc>>,
}

impl From<Message> for MessageWatcher {
    fn from(msg: Message) -> Self {
        MessageWatcher {
            message: msg,
            tries: 0,
            next_try_after: None,
        }
    }
}

pub struct MailQueue {
    pub(crate) mails: RwLock<VecDeque<MessageWatcher>>,
}

impl MailQueue {
    pub fn add_mail(&self, mail: Message) -> Result<(), std::sync::PoisonError<std::sync::RwLockWriteGuard<'_, std::collections::VecDeque<MessageWatcher>>>> {
        self.mails.write()?.push_back(mail.into());
        Ok(())
    }
    pub fn add_msg_watcher(&self, msg_watcher: MessageWatcher) -> Result<(), std::sync::PoisonError<std::sync::RwLockWriteGuard<'_, std::collections::VecDeque<MessageWatcher>>>> {
        self.mails.write()?.push_back(msg_watcher);
        Ok(())
    }
    pub fn get_next(&self) -> Result<Option<MessageWatcher>, std::sync::PoisonError<std::sync::RwLockWriteGuard<'_, std::collections::VecDeque<MessageWatcher>>>> {
        Ok(self.mails.write()?.pop_front())
    }
    pub async fn process_next(&self, settings: &Settings) -> Result<(), std::sync::PoisonError<std::sync::RwLockWriteGuard<'_, std::collections::VecDeque<MessageWatcher>>>> {
        let mail_watcher = self.get_next()?;

        match mail_watcher {
            Some(mut mail_watcher) => {
                if let Some(next_try_after) = mail_watcher.next_try_after {
                    let current = Utc::now();
                    if next_try_after.ge(&current) { //Not due
                        self.add_msg_watcher(mail_watcher)?;
                        return Ok(())
                    }
                }

                match send_mail(mail_watcher.message.clone(), settings).await {
                    Ok(_) => {},
                    Err(_) => {
                        //Do not re-add mail after third try:
                        if mail_watcher.tries < 3 {
                            if mail_watcher.tries == 0 {
                                mail_watcher.next_try_after = Some(Utc::now().add(Duration::minutes(1))); //First retry after 1 minute
                            } else if mail_watcher.tries == 1 {
                                mail_watcher.next_try_after = Some(Utc::now().add(Duration::minutes(10))); //Second retry after 10 minutes
                            } else if mail_watcher.tries == 2 {
                                mail_watcher.next_try_after = Some(Utc::now().add(Duration::minutes(600))); //Third retry after 10 hours
                            }
                            mail_watcher.tries = mail_watcher.tries + 1;
                            self.add_msg_watcher(mail_watcher)?;
                        }
                    }
                }
                Ok(())
            },
            None => Ok(())
        }
    }
    pub fn load_or_create_new() -> MailQueue {
        /*match std::fs::File::open("mail_queue.txt"){
            Ok(mut file) => {
                let mut contents = String::new();
                match file.read_to_string(&mut contents){
                    Ok(_) => {
                        match serde_json::from_str::<MailQueue>(&contents){
                            Ok(queue) => return queue,
                            Err(e) => {
                                panic!("Couldn't parse mail queue file: {}", e);
                            }
                        }
                    }
                    Err(e) => {
                        panic!("Couldn't read mail_queue.txt! {}", e);
                    }
                }
            }
            Err(e) => {
                warn!("Couldn't read mail queue from disk: {}. Creating new empty queue.", e);*/
        MailQueue {
            mails: RwLock::new(VecDeque::new())
        }/*
            }
        }*/
    }/*
    fn save_to_disk(&self) -> std::io::Result<()> {
        let serialized = serde_json::to_string(&self)?;
        let mut file = std::fs::File::create("mail_queue.txt")?;
        file.write_all(serialized.as_bytes())
    }*/
}