use std::convert::TryInto;

use serde::{Deserialize, Deserializer};

/// Warning: You have to use #\[serde(default)] on each struct field! Otherwise every missing field will be interpreted as null
#[derive(Debug, PartialEq)]
pub enum Patch<T> {
    //No value for field transmitted, treat as keep old
    Missing,
    //Null for field transmitted, treat as set to NULL
    Null,
    Value(T),
}

impl<T> Default for Patch<T> {
    fn default() -> Self {
        Patch::Missing
    }
}

impl<T> From<Option<T>> for Patch<T> {
    fn from(opt: Option<T>) -> Patch<T> {
        match opt {
            Some(v) => Patch::Value(v),
            None => Patch::Null,
        }
    }
}

impl<T> Into<Option<Option<T>>> for Patch<T> {
    /// Converts Patch into Option<Option<T>>, allowing null values
    ///
    /// * Patch::Missing results in None -> do not update field
    /// * Patch::Value(T) results in Some(Some(T)) -> update field with T
    /// * Patch::Null results in Some(None) -> set field to null
    fn into(self) -> Option<Option<T>> {
        match self {
            Patch::Missing => None,
            Patch::Null => Some(None),
            Patch::Value(v) => Some(Some(v))
        }
    }
}

impl<T> TryInto<Option<T>> for Patch<T> {
    type Error = ();

    /// Converts Patch into Option<T>, not allowing null values
    ///
    /// * Patch::Missing results in None -> do not update field
    /// * Patch::Value(T) results in Some(T) -> update field with T
    /// * Patch::Null results in an error -> trying to set field to null which isn't allowed
    fn try_into(self) -> Result<Option<T>, Self::Error> {
        match self {
            Patch::Missing => Ok(None),
            Patch::Null => Err(()),
            Patch::Value(v) => Ok(Some(v))
        }
    }
}

impl<'de, T> Deserialize<'de> for Patch<T>
    where
        T: Deserialize<'de>,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
        where
            D: Deserializer<'de>,
    {
        Option::deserialize(deserializer).map(Into::into)
    }
}
