use chrono::{DateTime, NaiveDateTime, ParseError, TimeZone, Utc};
use chrono_tz::Tz;
use rocket::State;

use crate::helper::session_cookies::model::SessionCookie;
use crate::Settings;

pub fn get_timezone(settings: &State<Settings>, cookie: Option<&SessionCookie>) -> Tz {
    let default_tz: Tz = match settings.application.default_timezone.parse() {
        Ok(tz) => tz,
        Err(e) => {
            panic!("Invalid configuration! Couldn't parse default_timezone: {}", e);
        }
    };
    if let Some(cookie) = cookie {
        match cookie.user.timezone.clone() {
            Some(tz) => match tz.parse() {
                Ok(tz) => tz,
                Err(e) => {
                    warn!("Unparsable timezone for user: {}, falling back to default timezone. Error is {}", cookie.user.id, e);
                    default_tz
                }
            },
            None => default_tz
        }
    } else {
        default_tz
    }
}

pub fn datetime_str_to_utc(settings: &State<Settings>, cookie: Option<&SessionCookie>, user_datetime: &str) -> Result<DateTime<Utc>, ParseError> {
    let tz = get_timezone(settings, cookie);

    let naive = NaiveDateTime::parse_from_str(user_datetime, "%Y-%m-%dT%H:%M")?;
    let local_datetime = tz.from_local_datetime(&naive).unwrap();
    Ok(local_datetime.with_timezone(&Utc))
}

pub fn utc_to_local_user_time(settings: &State<Settings>, cookie: Option<&SessionCookie>, utc: DateTime<Utc>) -> String {
    let tz = get_timezone(settings, cookie);
    let local_time = utc.with_timezone(&tz);
    local_time.format("%Y-%m-%dT%H:%M").to_string()
}

#[cfg(test)]
mod tests {
    use chrono::{DateTime, Local, TimeZone, Utc};
    use rocket::{Build, Rocket};

    use crate::database::model::users::User;
    use crate::helper::settings::Application;

    use super::*;

    fn get_test_settings() -> Settings {
        Settings {
            database: Default::default(),
            application: Application {
                url: "".to_string(),
                default_language: "".to_string(),
                fallback_language: "".to_string(),
                loglevel: "".to_string(),
                session_timeout: 0,
                upload_path: "".to_string(),
                max_login_attempts: 0,
                login_lock_duration: 0,
                name: "".to_string(),
                reset_password_token_valid_duration: 0,
                user_support_email: "".to_string(),
                new_member_default_role: "".to_string(),
                default_timezone: "Europe/Berlin".to_string(),
            },
            mail: Default::default(),
            api: Default::default(),
            billing: Default::default(),
        }
    }

    fn get_test_cookies() -> SessionCookie {
        SessionCookie {
            id: "".to_string(),
            expires: Utc::now(),
            user: User {
                id: Default::default(),
                password: None,
                email: "".to_string(),
                username: None,
                timezone: None,
            },
            member: None,
        }
    }

    #[test]
    pub fn test_utc_to_local_user_time() {
        let utcdate: DateTime<Utc> = Utc.ymd(2022, 1, 23).and_hms(12, 0, 0);

        let rocket = rocket::build().manage(get_test_settings());
        let res = utc_to_local_user_time(State::get(&rocket).unwrap(), Some(&get_test_cookies()), utcdate);

        assert_eq!(res, "2022-01-23T13:00");
    }

    #[test]
    pub fn test_datetime_str_to_utc_error() {
        let rocket = rocket::build().manage(get_test_settings());
        assert!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "2022-01-23").is_err());
    }

    #[test]
    pub fn test_datetime_str_to_utc_error2() {
        let rocket = rocket::build().manage(get_test_settings());
        assert!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "2022-01-23T").is_err());
    }

    #[test]
    pub fn test_datetime_str_to_utc_error3() {
        let rocket = rocket::build().manage(get_test_settings());
        assert!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "2022-01-23T15:00:00").is_err());
    }

    #[test]
    pub fn test_datetime_str_to_utc_error4() {
        let rocket = rocket::build().manage(get_test_settings());
        assert!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "2022-01-23T15").is_err());
    }

    #[test]
    pub fn test_datetime_str_to_utc_error5() {
        let rocket = rocket::build().manage(get_test_settings());
        assert!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "Black Mamba").is_err());
    }

    #[test]
    pub fn test_datetime_str_to_utc_correct() {
        let rocket = rocket::build().manage(get_test_settings());
        let utcdate: DateTime<Utc> = Utc.ymd(2022, 1, 23).and_hms(12, 0, 0);
        assert_eq!(datetime_str_to_utc(State::get(&rocket).unwrap(), Some(&get_test_cookies()), "2022-01-23T13:00").unwrap(), utcdate);
    }

    #[test]
    pub fn test_timezone() {
        let utcdate: DateTime<Utc> = Utc.ymd(2022, 1, 23).and_hms(12, 0, 0);
        let tz: Tz = "Europe/Berlin".parse().unwrap();
        let localdate1 = utcdate.with_timezone(&tz);

        let localdate2: DateTime<Local> = Local.ymd(2022, 1, 23).and_hms(13, 0, 0);

        print!("{}:{}", localdate1.to_string(), localdate2.to_string());
        assert_eq!(localdate1, localdate2);
    }

    #[test]
    pub fn test_timezone_to_utc() {
        let tz: Tz = "Europe/Berlin".parse().unwrap();
        let naive = NaiveDateTime::parse_from_str("2022-01-23T13:00", "%Y-%m-%dT%H:%M").unwrap();
        let local_datetime = tz.from_local_datetime(&naive).unwrap();
        let converted_to_utc: DateTime<Utc> = local_datetime.with_timezone(&Utc);
        let native_utc: DateTime<Utc> = Utc.ymd(2022, 1, 23).and_hms(12, 0, 0);

        print!("{}:{}", converted_to_utc.to_string(), native_utc.to_string());
        assert_eq!(converted_to_utc, native_utc);
    }
}