use chrono::{DateTime, Utc};
use diesel::sql_types::*;
use serde_json::Value;

use crate::logger::LogActions;

/// LogEntry to create a new log entry with specified action, affected_entity (optional), causer (optional) and optional details
/// Timestamp and log entry id are generated automatically
pub struct InsertableLogEntry {
    pub action: String,
    pub affected_entity: Option<uuid::Uuid>,
    pub causer: Option<uuid::Uuid>,
    pub details: Option<Value>
}

impl InsertableLogEntry{
    /// Create new LogEntry with given values
    pub fn new(action: LogActions, affected_entity: Option<uuid::Uuid>, causer: Option<uuid::Uuid>, details: Option<Value>) -> Self{
        InsertableLogEntry{
            action: action.value(),
            affected_entity,
            causer,
            details
        }
    }
}

#[derive(QueryableByName)]
pub struct LogEntry{
    #[sql_type = "Integer"]
    pub entry_id: i32,
    #[sql_type = "Timestamptz"]
    pub timestamp: DateTime<Utc>,
    #[sql_type = "Text"]
    pub action: String,
    #[sql_type = "Nullable<Uuid>"]
    pub affected_entity: Option<uuid::Uuid>,
    #[sql_type = "Nullable<Uuid>"]
    pub causer: Option<uuid::Uuid>,
    #[sql_type = "Nullable<Jsonb>"]
    pub details: Option<Value>
}

#[derive(Serialize)]
pub struct EventBillingStateApproveLogEntry{
    pub state: uuid::Uuid
}

#[cfg(test)]
mod tests {
    use serde_json::json;

    use super::*;

    #[test]
    pub fn test_conversion() {
        let test =  EventBillingStateApproveLogEntry{
            state: uuid::Uuid::parse_str("00000000-0000-0000-0000-000000001234").unwrap()
        };

        let test3 : Value = serde_json::to_value(test).unwrap();

        let comp = json!({"state": "00000000-0000-0000-0000-000000001234"});
        assert_eq!(test3, comp);
    }
}