use crate::helper::session_cookies::model::SessionCookie;
use rocket::State;
use crate::helper::settings::Settings;
use rocket::serde::json::Json;
use crate::modules::api::model::api_outcome::{ApiErrorWrapper, ApiError};
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::database::controller::entities::generate_entity;
use crate::database::model::events::{EventUnitPosition, EventUnitVehiclePosition};
use crate::database::controller::events::{add_event_unit_position, add_vehicle_position};
use crate::helper::translate_diesel_error::translate_diesel;

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct CreatePositionData{
    pub entity_id: Option<String>,
    pub name: String,
    pub description: Option<String>,
    pub requirements: String,
}

#[post("/api/events/units/positions", format = "json", data = "<create_position_data>")]
pub fn create_event_unit_position(
    settings: &State<Settings>,
    cookie: SessionCookie,
    create_position_data: Json<CreatePositionData>,
) -> Result<Json<EventUnitPosition>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Einsatzpositionen anzulegen!".to_string()).to_wrapper(),
        ));
    }

    let create_position_data = create_position_data.into_inner();
    let entity_id = match generate_entity(settings){
        Ok(entity) => entity,
        Err(_e) => return Err(Json(ApiError::new(500, "Konnte keine neue Entität anlegen.".to_string()).to_wrapper())),
    };

    let requirements = match serde_json::from_str(&create_position_data.requirements){
        Ok(req) => req,
        Err(e) => {
            warn!("Couldn't parse requirements: {}", e);
            return Err(Json(ApiError::new(400, format!("Couldn't parse requirements: {}", e)).to_wrapper()))
        }
    };
    let event_position = EventUnitPosition{
        entity_id,
        name: create_position_data.name,
        description: create_position_data.description,
        requirements
    };

    match add_event_unit_position(settings, event_position){
        Ok(eup) => Ok(Json(eup)),
        Err(e) =>  Err(translate_diesel(e))
    }

}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct CreateVehiclePositionData{
    pub id: Option<String>,
    pub name: String,
    pub description: Option<String>,
    pub required_vehicle_category: String,
    pub template_id: String,
}

#[post("/api/events/units/vehicle_positions", format = "json", data = "<create_vehicle_position_data>")]
pub fn create_event_unit_vehicle_position(
    settings: &State<Settings>,
    cookie: SessionCookie,
    create_vehicle_position_data: Json<CreateVehiclePositionData>,
) -> Result<Json<EventUnitVehiclePosition>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Einsatzpositionen anzulegen!".to_string()).to_wrapper(),
        ));
    }

    let cvpd = create_vehicle_position_data.into_inner();
    let entity_id = match generate_entity(settings){
        Ok(entity) => entity,
        Err(_e) => return Err(Json(ApiError::new(500, "Konnte keine neue Entität anlegen.".to_string()).to_wrapper())),
    };

    let pos = EventUnitVehiclePosition{
        entity_id,
        name: cvpd.name,
        description: cvpd.description,
        required_vehicle_category: parse_uuid_string(cvpd.required_vehicle_category)?,
        template_id: parse_uuid_string(cvpd.template_id)?,
    };

    match add_vehicle_position(settings, pos){
        Ok(pos) => Ok(Json(pos)),
        Err(e) => Err(translate_diesel(e))
    }
}