use rocket::State;
use crate::helper::settings::Settings;
use crate::helper::session_cookies::model::SessionCookie;
use rocket::serde::json::Json;
use crate::database::model::events::EventUnitPosition;
use crate::modules::api::model::api_outcome::{ApiErrorWrapper, ApiError};
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::database::controller::events::{get_event_unit_positions, get_event_unit_positions_count};
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::event_management::check_position_requirements::{check_position_requirements, RequirementParserError};

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct EventUnitPositionList{
    pub(crate) positions: Vec<EventUnitPosition>,
    pub(crate) total_position_count: i64,
}

#[get("/api/events/units/positions?<limit>&<offset>&<q>", format = "json")]
pub fn read_event_unit_position(
    settings: &State<Settings>,
    cookie: SessionCookie,
    limit: Option<i64>,
    offset: Option<i64>,
    q: Option<String>
) -> Result<Json<EventUnitPositionList>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Einsatzpositionen abzurufen!".to_string()).to_wrapper(),
        ));
    }

    let limit = match limit{
        Some(limit) => limit,
        None => settings.api.default_pagination_limit,
    };
    let offset = match offset{
        Some(offset) => offset,
        None => 0,
    };

    let positions = match get_event_unit_positions(settings, limit, offset, q){
        Ok(pos) => pos,
        Err(e) => return Err(translate_diesel(e))
    };

    let total_position_count = match get_event_unit_positions_count(settings){
        Ok(count) => count,
        Err(e) => return Err(translate_diesel(e))
    };

    Ok(Json(EventUnitPositionList{
        positions,
        total_position_count
    }))

}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct PositionCheckResult{
    pub(crate) requirements_fulfilled: bool,
}


#[get("/api/events/units/positions/<position_id>/check_requirements?<target_to_check>", format = "json")]
pub fn get_check_position_requirements(
    settings: &State<Settings>,
    cookie: SessionCookie,
    position_id: String,
    target_to_check: String,
) -> Result<Json<PositionCheckResult>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    match check_position_requirements(settings, parse_uuid_string(position_id)?, parse_uuid_string(target_to_check)?){
        Ok(res) => Ok(Json(PositionCheckResult{
            requirements_fulfilled: res
        })),
        Err(e) => {
            match e{
                RequirementParserError::Database(e) => {
                    Err(translate_diesel(e))
                }
                RequirementParserError::Parser(e) => {
                    error!("Couldn't parse requirements: {}", e.as_str());
                    Err(Json(ApiError::new(400, "invalid json for requirements".to_string()).to_wrapper()))
                }
            }
        }
    }
}