use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::entities::generate_entity;
use crate::database::controller::events::{add_event_unit_template, add_position_to_template, RawEventUnitTemplatePosition};
use crate::database::model::events::EventUnitTemplate;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct CreateTemplateData {
    pub entity_id: Option<String>,
    pub name: String,
    pub description: Option<String>,
}

#[post("/api/events/units/templates", format = "json", data = "<create_template_data>")]
pub fn create_event_unit_template(
    settings: &State<Settings>,
    cookie: SessionCookie,
    create_template_data: Json<CreateTemplateData>,
) -> Result<Json<EventUnitTemplate>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Vorlagen anzulegen!".to_string()).to_wrapper(),
        ));
    }

    let create_template_data = create_template_data.into_inner();
    let entity_id = match generate_entity(settings){
        Ok(entity) => entity,
        Err(_e) => return Err(Json(ApiError::new(500, "Konnte keine neue Entität anlegen.".to_string()).to_wrapper())),
    };

    let event_position = EventUnitTemplate{
        entity_id,
        name: create_template_data.name,
        description: create_template_data.description,
    };

    match add_event_unit_template(settings, event_position) {
        Ok(eup) => Ok(Json(eup)),
        Err(e) => Err(translate_diesel(e))
    }
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct PutPositionInTemplateData {
    pub template_id: String,
    pub position_entity_id: String,
    pub tag: Option<String>,
    pub num: String,
}


#[post("/api/events/units/templates/<template_id>/positions/<position_id>", format = "json", data = "<put_position_in_template_data>")]
pub fn put_position_in_template(
    settings: &State<Settings>,
    cookie: SessionCookie,
    put_position_in_template_data: Json<PutPositionInTemplateData>,
    template_id: String,
    position_id: String,
) -> Result<(), Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Vorlagen zu verändern!".to_string()).to_wrapper(),
        ));
    }

    let data = put_position_in_template_data.into_inner();

    if data.template_id != template_id || data.position_entity_id != position_id {
        return Err(Json(
            ApiError::new(400, "template_id or position_entity_id in URI doesn't match POST data!".to_string()).to_wrapper(),
        ));
    }

    let entity_id = match generate_entity(settings) {
        Ok(entity) => entity,
        Err(_) => {
            return Err(Json(ApiError::new(500, "Konnte keine neue Entität anlegen.".to_string()).to_wrapper()))
        },
    };

    let num = match data.num.parse() {
        Ok(num) => num,
        Err(e) => {
            warn!("Wrong data type for num while putting position in template: {}", e);
            return Err(Json(ApiError::new(400, "Wrong data type for num.".to_string()).to_wrapper()))
        }
    };

    let res = RawEventUnitTemplatePosition {
        position_entity_id: parse_uuid_string(position_id)?,
        template_id: parse_uuid_string(template_id)?,
        position_template_id: entity_id,
        num,
        tag: data.tag,
    };

    match add_position_to_template(settings, res) {
        Ok(()) => Ok(()),
        Err(e) => Err(translate_diesel(e))
    }
}