use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::entities::remove_entity;
use crate::database::controller::events::remove_position_from_template;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[delete("/api/events/units/templates", format = "json", data = "<template_list>")]
pub fn delete_templates(
    settings: &State<Settings>,
    cookie: SessionCookie,
    template_list: Json<Vec<uuid::Uuid>>,
) -> Result<(), Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Vorlage zu löschen!".to_string()).to_wrapper(),
        ));
    }

    let template_list = template_list.into_inner();

    for template in template_list {
        match remove_entity(settings, template){
            Ok(_) => {}
            Err(e) => return Err(translate_diesel(e)),
        }
    }

    Ok(())
}

#[delete("/api/events/units/templates/<template_id>/positions/<position_id>/<position_template_id>", format = "json")]
pub fn delete_positions_from_template(
    settings: &State<Settings>,
    cookie: SessionCookie,
    template_id: String,
    position_id: String,
    position_template_id: String,
) -> Result<(), Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Vorlage zu ändern!".to_string()).to_wrapper(),
        ));
    }

    match remove_position_from_template(settings, parse_uuid_string(template_id)?, parse_uuid_string(position_id)?, parse_uuid_string(position_template_id)?) {
        Ok(_) => {}
        Err(e) => return Err(translate_diesel(e)),
    }

    Ok(())
}
