use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::events::{EventUnitTemplatePosition, get_event_unit_positions_for_template, get_event_unit_templates, get_event_unit_templates_count};
use crate::database::controller::events::templates::vehicle_positions::get_eu_vehicle_positions_for_template;
use crate::database::model::events::{EventUnitTemplate, EventUnitVehiclePosition};
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct EventUnitTemplateList {
    pub(crate) templates: Vec<EventUnitTemplate>,
    pub(crate) total_template_count: i64,
}

#[get("/api/events/units/templates?<limit>&<offset>&<q>", format = "json")]
pub fn read_event_unit_templates(
    settings: &State<Settings>,
    cookie: SessionCookie,
    limit: Option<i64>,
    offset: Option<i64>,
    q: Option<String>
) -> Result<Json<EventUnitTemplateList>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Vorlagen abzurufen!".to_string()).to_wrapper(),
        ));
    }

    let limit = match limit{
        Some(limit) => limit,
        None => settings.api.default_pagination_limit,
    };
    let offset = match offset{
        Some(offset) => offset,
        None => 0,
    };

    let templates = match get_event_unit_templates(settings, limit, offset, q){
        Ok(pos) => pos,
        Err(e) => return Err(translate_diesel(e))
    };

    let total_template_count = match get_event_unit_templates_count(settings){
        Ok(count) => count,
        Err(e) => return Err(translate_diesel(e))
    };

    Ok(Json(EventUnitTemplateList{
        templates,
        total_template_count
    }))

}

#[get("/api/events/units/templates/<template_id>/positions", format = "json")]
pub fn read_event_unit_template_positions(
    settings: &State<Settings>,
    cookie: SessionCookie,
    template_id: String
) -> Result<Json<Vec<EventUnitTemplatePosition>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Vorlagenpositionen abzurufen!".to_string()).to_wrapper(),
        ));
    }

    let positions = match get_event_unit_positions_for_template(settings, parse_uuid_string(template_id)?){
        Ok(pos) => pos,
        Err(e) => return Err(translate_diesel(e))
    };

    Ok(Json(positions))
}


#[get("/api/events/units/templates/<template_id>/vehicle_positions", format = "json")]
pub fn read_event_unit_template_vehicle_positions(
    settings: &State<Settings>,
    cookie: SessionCookie,
    template_id: String
) -> Result<Json<Vec<EventUnitVehiclePosition>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Fahrzeugpositionen abzurufen!".to_string()).to_wrapper(),
        ));
    }

    let positions = match get_eu_vehicle_positions_for_template(settings, parse_uuid_string(template_id)?) {
        Ok(pos) => pos,
        Err(e) => return Err(translate_diesel(e))
    };

    Ok(Json(positions))
}