use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::entities::generate_entity;
use crate::database::controller::events::{add_instance, add_position_instances_for_instance};
use crate::database::controller::events::instances::instance_positions::{create, RawPositionInstanceInsert};
use crate::database::controller::events::instances::instances::RawEventUnitInstance;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::time::datetime_str_to_utc;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::events::instances::read::PositionInstance;
use crate::modules::api::member_management::controller::parser::{
    parse_member_cookie, parse_uuid_string,
};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct CreateInstanceData {
    pub template_id: String,
    pub name: String,
    pub event_id: uuid::Uuid,
    pub planned_start_time: Option<String>,
    pub planned_end_time: Option<String>,
}

#[post(
"/api/events/<event_id>/instances",
format = "json",
data = "<create_instance_data>"
)]
pub fn create_instance(
    settings: &State<Settings>,
    cookie: SessionCookie,
    create_instance_data: Json<CreateInstanceData>,
    event_id: String,
) -> Result<Json<RawEventUnitInstance>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member.clone())?;
    if !caller
        .has_permission(crate::permissions::modules::event_management::events::EDIT.to_string())
    {
        return Err(Json(
            ApiError::new(
                403,
                "Keine Berechtigung Einsätze zu bearbeiten!".to_string(),
            )
                .to_wrapper(),
        ));
    }
    let cid = create_instance_data.into_inner();

    if parse_uuid_string(event_id.clone())? != cid.event_id {
        return Err(Json(
            ApiError::new(
                400,
                "Two different event_ids in body and parameter!".to_string(),
            )
                .to_wrapper(),
        ));
    }
    let entity_id = match generate_entity(settings) {
        Ok(entity) => entity,
        Err(_e) => {
            return Err(Json(
                ApiError::new(500, "Konnte keine neue Entität anlegen.".to_string()).to_wrapper(),
            ))
        }
    };

    let planned_start_time = match cid.planned_start_time {
        Some(dt) => match datetime_str_to_utc(settings, Some(&cookie), &dt) {
            Ok(dt) => Some(dt),
            Err(e) => {
                warn!("Couldn't parse planned start time as DateTime: {}", e);
                return Err(Json(
                    ApiError::new(400, "Couldn't parse planned start time!".to_string())
                        .to_wrapper(),
                ));
            }
        },
        None => None,
    };

    let planned_end_time = match cid.planned_end_time {
        Some(dt) => match datetime_str_to_utc(settings, Some(&cookie), &dt) {
            Ok(dt) => Some(dt),
            Err(e) => {
                warn!("Couldn't parse planned end time as DateTime: {}", e);
                return Err(Json(
                    ApiError::new(400, "Couldn't parse planned end time!".to_string()).to_wrapper(),
                ));
            }
        },
        None => None,
    };

    let instance = RawEventUnitInstance {
        instance_id: entity_id,
        template_id: parse_uuid_string(cid.template_id)?,
        name: cid.name,
        event_id: parse_uuid_string(event_id)?,
        planned_start_time,
        planned_end_time,
        real_start_time: None,
        real_end_time: None,
        billing_rate_id: None,
        billing_state_id: None,
        billing_state_author: None
    };

    match add_position_instances_for_instance(settings, instance.instance_id, instance.template_id)
    {
        Ok(_) => {}
        Err(e) => return Err(translate_diesel(e)),
    }

    return match add_instance(settings, instance) {
        Ok(instance) => Ok(Json(instance)),
        Err(e) => Err(translate_diesel(e)),
    }
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct CreatePositionInstanceData {
    pub(crate) instance_id: uuid::Uuid,
    pub(crate) position_id: uuid::Uuid,
    pub(crate) taken_by: Option<uuid::Uuid>,
    pub(crate) real_start_time: Option<String>,
    pub(crate) real_end_time: Option<String>,
}

#[post(
"/api/events/instances/<instance_id>/positions/<position_id>",
format = "json", data = "<create_position_instance_data>"
)]
pub fn add_position_instance(
    settings: &State<Settings>,
    cookie: SessionCookie,
    instance_id: String,
    position_id: String,
    create_position_instance_data: Json<CreatePositionInstanceData>,
) -> Result<Json<PositionInstance>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member.clone())?;
    //TODO: allow member_responsible for event to bypass this permission
    if !caller
        .has_permission(crate::permissions::modules::event_management::events::EDIT.to_string())
    {
        return Err(Json(
            ApiError::new(
                403,
                "Keine Berechtigung Einsätze zu bearbeiten!".to_string(),
            )
                .to_wrapper(),
        ));
    }

    let create_position_instance_data = create_position_instance_data.into_inner();

    if parse_uuid_string(instance_id)? != create_position_instance_data.instance_id {
        return Err(Json(
            ApiError::new(
                400,
                "Two different instance_ids in body and parameter!".to_string(),
            )
                .to_wrapper(),
        ));
    }

    if parse_uuid_string(position_id)? != create_position_instance_data.position_id {
        return Err(Json(
            ApiError::new(
                400,
                "Two different instance_ids in body and parameter!".to_string(),
            )
                .to_wrapper(),
        ));
    }
    //TODO: require extra permissions to set real start time & real end time
    let real_start_time = match create_position_instance_data.real_start_time {
        Some(dt) => match datetime_str_to_utc(settings, Some(&cookie), &dt) {
            Ok(dt) => Some(dt),
            Err(e) => {
                warn!("Couldn't parse real start time as DateTime: {}", e);
                return Err(Json(
                    ApiError::new(400, "Couldn't parse real start time!".to_string())
                        .to_wrapper(),
                ));
            }
        },
        None => None,
    };
    let real_end_time = match create_position_instance_data.real_end_time {
        Some(dt) => match datetime_str_to_utc(settings, Some(&cookie), &dt) {
            Ok(end) => {
                if let Some(start) = real_start_time {
                    if start.ge(&end) {
                        return Err(Json(ApiError::new(400, "Start time can't be later than the end time!".to_string()).to_wrapper()))
                    }
                }
                Some(end)
            },
            Err(e) => {
                warn!("Couldn't parse real end time as DateTime: {}", e);
                return Err(Json(
                    ApiError::new(400, "Couldn't parse real end time!".to_string())
                        .to_wrapper(),
                ));
            }
        },
        None => None,
    };

    let data = RawPositionInstanceInsert {
        position_instance_id: None,
        instance_id: create_position_instance_data.instance_id,
        position_id: create_position_instance_data.position_id,
        taken_by: create_position_instance_data.taken_by,
        real_start_time,
        real_end_time,
    };

    match create(settings, data) {
        Ok(res) => {
            match PositionInstance::from_raw(&res, settings, &cookie) {
                Ok(pi) => Ok(Json(pi)),
                Err(_) => return Err(Json(ApiError::new(500, "Couldn't retrieve position instance!".to_string()).to_wrapper())),
            }
        },
        Err(e) => Err(translate_diesel(e))
    }
}
