use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::events::{get_eu_position, get_instance_vehicle_positions, get_position_instances};
use crate::database::controller::events::instances::instance_positions::RawPositionInstance;
use crate::database::controller::events::instances::instances::{get_instances, RawEventUnitInstance};
use crate::database::model::events::EventUnitInstanceVehiclePosition;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::time::utc_to_local_user_time;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};
use crate::modules::api::personnel_billing::calculation::calculate_hours;

/// This struct contains RawEventUnitInstance data but converts all UTC DateTimes to local time
#[derive(Queryable, Clone, Deserialize, Serialize, )]
pub struct EventUnitInstance {
    pub instance_id: uuid::Uuid,
    pub template_id: uuid::Uuid,
    pub name: String,
    pub event_id: uuid::Uuid,
    pub planned_start_time: Option<String>,
    pub planned_end_time: Option<String>,
    pub real_start_time: Option<String>,
    pub real_end_time: Option<String>,
    pub billing_rate_id: Option<uuid::Uuid>,
    pub billing_state_id: Option<uuid::Uuid>,
}

impl EventUnitInstance {
    /// Convert RawEventUnitInstance to EventUnitInstance by converting Utc DateTimes to timestamps with user specified timezone
    pub fn from_raw(raw: RawEventUnitInstance, settings: &State<Settings>, cookie: &SessionCookie) -> EventUnitInstance {
        let planned_start_time = match raw.planned_start_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };
        let planned_end_time = match raw.planned_end_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };
        let real_start_time = match raw.real_start_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };
        let real_end_time = match raw.real_end_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };

        EventUnitInstance {
            instance_id: raw.instance_id,
            template_id: raw.template_id,
            name: raw.name,
            event_id: raw.event_id,
            planned_start_time,
            planned_end_time,
            real_start_time,
            real_end_time,
            billing_rate_id: raw.billing_rate_id,
            billing_state_id: raw.billing_state_id,
        }
    }
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct PositionInstance {
    pub(crate) position_instance_id: uuid::Uuid,
    pub(crate) instance_id: uuid::Uuid,
    pub(crate) position_id: uuid::Uuid,
    pub(crate) position_name: String,
    pub(crate) position_description: Option<String>,
    pub(crate) position_requirements: Option<serde_json::Value>,
    pub(crate) taken_by: Option<uuid::Uuid>,
    pub(crate) real_start_time: Option<String>,
    pub(crate) real_end_time: Option<String>,
    pub(crate) billable: Option<bool>,
    pub(crate) fulfilled_time: Option<i32>,
}

impl PositionInstance {
    /// Convert RawPositionInstance to PositionInstance by converting Utc DateTimes to timestamps with user specified timezone
    pub fn from_raw(raw: &RawPositionInstance, settings: &State<Settings>, cookie: &SessionCookie) -> Result<PositionInstance, ()> {
        let real_start_time = match raw.real_start_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };
        let real_end_time = match raw.real_end_time {
            Some(dt) => Some(utc_to_local_user_time(settings, Some(cookie), dt)),
            None => None
        };
        let position_data = match get_eu_position(settings, raw.position_id) {
            Ok(position) => position,
            Err(_e) => { //TODO: distinguish personal & vehicle positions
                return Err(())
            }
        };
        //Calculate hours
        let other_fulfilled_time = if let Some(start) = raw.real_start_time {
            if let Some(end) = raw.real_end_time {
                calculate_hours(start, end)
            } else {
                None
            }
        } else {
            None
        };

        Ok(PositionInstance {
            position_instance_id: raw.position_instance_id,
            instance_id: raw.instance_id,
            position_id: raw.position_id,
            position_name: position_data.name,
            position_description: position_data.description,
            position_requirements: position_data.requirements,
            taken_by: raw.taken_by,
            real_start_time,
            real_end_time,
            billable: raw.billable,
            fulfilled_time: other_fulfilled_time,
        })
    }
}

#[get("/api/events/<event_id>/instances", format = "json", rank = 1)]
pub fn read_instances(
    settings: &State<Settings>,
    cookie: SessionCookie,
    event_id: String,
) -> Result<Json<Vec<EventUnitInstance>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member.clone())?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::VIEW.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Einsätze abzurufen!".to_string()).to_wrapper(),
        ));
    }

    let pos = match get_instances(settings, parse_uuid_string(event_id)?) {
        Ok(pos) => pos,
        Err(e) => return Err(translate_diesel(e))
    };

    let mut res: Vec<EventUnitInstance> = vec![];
    for instance in pos {
        res.push(EventUnitInstance::from_raw(instance, &settings, &cookie));
    }

    Ok(Json(res))
}

/// Get a list of all position instances for specific instance_id
#[get("/api/events/instances/<instance_id>/positions", format = "json", rank = 1)]
pub fn read_positions_for_instance(settings: &State<Settings>, cookie: SessionCookie, instance_id: String) -> Result<Json<Vec<PositionInstance>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member.clone())?;
    //TODO: allow member_responsible for event to bypass this permission
    if !caller.has_permission(crate::permissions::modules::event_management::events::VIEW.to_string()) {
        return Err(Json(ApiError::new(403, "Keine Berechtigung Einsätze abzurufen!".to_string()).to_wrapper()));
    }

    match get_position_instances(settings, parse_uuid_string(instance_id)?) {
        Ok(pos) => {
            let mut res = vec![];
            for position in pos {
                //try to convert RawPositionInstance to PositionInstance.
                match PositionInstance::from_raw(&position, settings, &cookie) {
                    Ok(pi) => res.push(pi),
                    Err(_) => {
                        //return Err(Json(ApiError::new(500, "Couldn't retrieve position instance".to_string()).to_wrapper()))
                        //TODO: distinguish vehicle & personnel positions
                    }
                }
            }
            Ok(Json(res))
        },
        Err(e) => return Err(translate_diesel(e))
    }
}

#[get("/api/events/instances/<instance_id>/vehicle_positions", format = "json", rank = 1)]
pub fn read_vehicle_positions_for_instance(
    settings: &State<Settings>,
    cookie: SessionCookie,
    instance_id: String,
) -> Result<Json<Vec<EventUnitInstanceVehiclePosition>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::events::VIEW.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung Einsätze abzurufen!".to_string()).to_wrapper(),
        ));
    }

    match get_instance_vehicle_positions(settings, parse_uuid_string(instance_id)?){
        Ok(pos) => Ok(Json(pos)),
        Err(e) => return Err(translate_diesel(e))
    }
}