use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::event_requests::{get_event_request_count, get_event_requests};
use crate::database::model::event_requests::EventRequest;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::parse_member_cookie;
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct EventRequestList {
    pub(crate) requests: Vec<EventRequest>,
    pub(crate) total_request_count: i64,
}

/// get list of all event requests
///
/// Arguments:
/// * settings: [Settings] struct managed by rocket
/// * cookie: [SessionCookie]
/// * limit: max row count, optional
/// * offset: skip first x rows, optional
/// * state: representing request state:
///     * 0: unknown
///     * 1: new (new request from extern)
///     * 3: validated (request has been checked)
///     * 5: published (request published)
///     * 7: adopted (some group admin adopted request)
#[get("/api/event_requests?<limit>&<offset>&<state>", format = "json")]
pub fn read_event_requests(
    settings: &State<Settings>,
    cookie: SessionCookie,
    limit: Option<i64>,
    offset: Option<i64>,
    state: Option<i16>,
) -> Result<Json<EventRequestList>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::event_management::requests::VIEW.to_string()) {
        return Err(Json(ApiError::new(403, "Keine Berechtigung Einsatzanfragen abzurufen!".to_string()).to_wrapper()))
    }

    let limit = match limit{
        Some(limit) => limit,
        None => settings.api.default_pagination_limit,
    };
    let offset = match offset{
        Some(offset) => offset,
        None => 0,
    };

    let requests = match get_event_requests(settings, limit, offset, state){
        Ok(events) => events,
        Err(e) => return Err(translate_diesel(e)),
    };
    let total_request_count = match get_event_request_count(settings, state){
        Ok(count) => count,
        Err(e) => return Err(translate_diesel(e)),
    };

    Ok(Json(EventRequestList{
        requests,
        total_request_count
    }))
}