use std::sync::Arc;

use chrono::NaiveDateTime;
use lettre::Message;
use lettre::message::header;
use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::api_communication_targets::get_member_email_addresses;
use crate::database::controller::event_requests::{change_event_request, get_event_request};
use crate::database::controller::members::get_member_by_uuid;
use crate::database::controller::organisers::get_organiser;
use crate::database::controller::permissions::get_members_with_permission;
use crate::database::model::event_requests::EventRequest;
use crate::database::model::organisers::Organiser;
use crate::helper::mail_queue::queue::MailQueue;
use crate::helper::mail_templates::MailTemplates;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::events::requests::create::RequestData;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_option_uuid, parse_uuid};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[put("/api/event_requests/<request_id>", format = "json", data = "<create_request_data>")]
pub fn update_event_request(
    settings: &State<Settings>,
    cookie: SessionCookie,
    create_request_data: Json<RequestData>,
    request_id: String,
    mt: &State<MailTemplates>, mq: &State<Arc<MailQueue>>
) -> Result<Json<EventRequest>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission(crate::permissions::modules::event_management::requests::EDIT.to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Einsatzanfragen zu verändern!".to_string()).to_wrapper(),
        ));
    }

    let entity_id = parse_uuid(&request_id)?;

    let old_data = match get_event_request(settings, entity_id){
        Ok(opt) => match opt{
            Some(old) => old,
            None => return Err(Json(ApiError::new(404, "Didn't found event request with specified id".to_string()).to_wrapper()))
        },
        Err(e) => return Err(translate_diesel(e))
    };

    let crd = create_request_data.into_inner();

    let risk_potential : Option<i32> = match crd.risk_potential{
        Some(rp) => match rp.parse() {
            Ok(rp) => Some(rp),
            Err(e) => {
                warn!("Couldn't parse risk potential for new request: {}", e);
                return Err(Json(ApiError::new(500, "couldn't parse risk potential".to_string()).to_wrapper()))
            }
        },
        None => None,
    };

    let expected_attendees : Option<i32> = match crd.expected_attendees{
        Some(ea) => match ea.parse() {
            Ok(ea) => Some(ea),
            Err(e) => {
                warn!("Couldn't parse expected attendees for new request: {}", e);
                return Err(Json(ApiError::new(500, "couldn't parse expected attendees".to_string()).to_wrapper()))
            }
        },
        None => None,
    };

    let received = match crd.received{
        Some(mut rec) => {
            rec += "T00:00";
            match NaiveDateTime::parse_from_str(&rec, "%Y-%m-%dT%H:%M") {
                Ok(rec) => rec,
                Err(e) => {
                    error!("Couldn't parse start datetime: {}", e);
                    return Err(Json(ApiError::new(400, "Das eingegebene Datum konnte nicht verarbeitet werden.".to_string()).to_wrapper()))
                }
            }
        },
        None => {
            old_data.received
        }
    };

    let state = match crd.state{
        Some(state) => match state.parse::<i16>() {
            Ok(state) => {
                if state != 1 && state != 0 && state != 3 && state != 5 && state != 7{
                    warn!("Unknown state: {}", state);
                    return Err(Json(ApiError::new(400, "Unknown state number.".to_string()).to_wrapper()))
                }else{
                    state
                }
            },
            Err(e) => {
                error!("Couldn't parse state: {}", e);
                return Err(Json(ApiError::new(400, "Couldn't parse state.".to_string()).to_wrapper()))
            }
        },
        None => 0
    };

    let input = EventRequest{
        entity_id,
        name: crd.name,
        timescale: crd.timescale,
        site: crd.site,
        organiser_id: parse_option_uuid(crd.organiser_id)?,
        etype: parse_option_uuid(crd.etype)?,
        related_group: parse_option_uuid(crd.related_group)?,
        contact_on_site_name: crd.contact_on_site_name,
        contact_on_site_phone: crd.contact_on_site_phone,
        specials: crd.specials,
        requested: crd.requested,
        expected_attendees,
        risk_potential,
        state,
        received,
    };

    match change_event_request(&settings, input){
        Ok(request) => {
            if state == 5 && old_data.state != 5{
                debug!("State 5 is new, sending emails!");
                send_event_request_published_emails(mt, mq, settings, request.clone())
            }else{
                debug!("State is: {}, old state was: {}", state, old_data.state);
            }
            Ok(Json(request))
        },
        Err(e) => Err(translate_diesel(e))
    }
}

#[derive(Serialize)]
struct NewEventRequestPublishedEmail{
    firstname: String,
    request: EventRequest,
    frontpage: String,
    support_email: String,
    organiser: Option<Organiser>
}

fn send_event_request_published_emails(mt: &State<MailTemplates>, mq: &State<Arc<MailQueue>>, settings: &State<Settings>, request: EventRequest) {
    let receivers = match get_members_with_permission(settings, crate::permissions::modules::event_management::requests::assignments::ASSIGN){
        Ok(rec) => rec,
        Err(e) => {
            error!("Couldn't get email receiver list: {}", e);
            return;
        }
    };

    let organiser = match request.clone().organiser_id{
        Some(organiser) => {
            match get_organiser(settings, organiser){
                Ok(organiser) => Some(organiser),
                Err(e) => {
                    error!("Couldn't get organiser for new event request to send email: {}", e);
                    None
                }
            }
        },
        None => None
    };

    for receiver in receivers {
        let mut mail = Message::builder()
            .from(settings.mail.from.clone().parse().unwrap())
            .reply_to(settings.mail.reply_to.clone().parse().unwrap())
            .subject("Einsatz Online - Passwort Zurücksetzen");

        match get_member_email_addresses(settings, receiver) {
            Ok(emails) => {
                for email in emails {
                    match email.parse() {
                        Ok(email) => {
                            mail = mail.to(email);
                        },
                        Err(e) => {
                            error!("Couldn't parse email address {}: {}", email, e);
                        }
                    }
                }
            },
            Err(e) => {
                error!("Couldn't get email addresses for member: {}", e);
                continue
            }
        };
        let member = match get_member_by_uuid(receiver, settings) {
            Some(member) => member,
            None => {
                error!("No member found for id {}", receiver);
                continue;
            }
        };

        let nerpe = NewEventRequestPublishedEmail {
            firstname: member.firstname,
            request: request.clone(),
            frontpage: settings.application.url.clone(),
            support_email: settings.application.user_support_email.clone(),
            organiser: organiser.clone(),
        };
        let mail = mail.header(header::ContentType::TEXT_PLAIN).body(match mt.registry.render("new_event_request_published-de", &nerpe) {
            Ok(body) => body,
            Err(e) => {
                error!("Couldn't render email template: {}", e);
                return
            },
        }).unwrap();

        mq.add_mail(mail);
    }

}