use std::convert::TryFrom;

use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::api_members::{get_member_list, get_member_list_count, get_member_search_result, get_member_search_result_by_name, MemberListEntryWithPermissions};
use crate::database::controller::api_members::MemberListOrder;
use crate::database::controller::groups::GroupMembership;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

#[derive(Serialize, Deserialize, Queryable, Clone)]
pub struct MemberSearchResult {
    pub(crate) entity_id: uuid::Uuid,
    pub(crate) firstname: String,
    pub(crate) lastname: String,
    pub(crate) readable: bool,
    pub(crate) groups: Vec<GroupMembership>,
}

#[derive(Serialize, Deserialize, Queryable, Clone)]
pub struct MemberSearch {
    pub(crate) members: Vec<MemberSearchResult>,
}

#[get("/api/members/<entity_id>", format = "json", rank = 1)]
pub fn api_members_get(
    cookie: SessionCookie,
    settings: &State<Settings>,
    entity_id: String,
) -> Result<Json<MemberSearchResult>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission("modules.member_management.search".to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Mitglieder zu suchen!".to_string())
                .to_wrapper(),
        ));
    };

    let entity_id = parse_uuid_string(entity_id)?;
    match get_member_search_result(settings, entity_id, caller.entity_id) {
        Ok(result) => Ok(Json(result)),
        Err(e) => match e {
            diesel::result::Error::NotFound => Err(Json(
                ApiError::new(
                    404,
                    "Das angeforderte Mitglied konnte nicht gefunden werden!".to_string(),
                )
                .to_wrapper(),
            )),
            _ => Err(Json(
                ApiError::new(500, "Es ist ein Datenbankfehler aufgetreten!".to_string())
                    .to_wrapper(),
            )),
        },
    }
}

#[get("/api/members?<name>", format = "json", rank = 2)]
pub fn api_members_search_name(
    cookie: SessionCookie,
    settings: &State<Settings>,
    name: String,
) -> Result<Json<MemberSearch>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission("modules.member_management.search".to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Mitglieder zu suchen!".to_string())
                .to_wrapper(),
        ));
    };

    match get_member_search_result_by_name(settings, name, caller.entity_id) {
        Ok(result) => Ok(Json(MemberSearch { members: result })),
        Err(e) => match e {
            diesel::result::Error::NotFound => Err(Json(
                ApiError::new(
                    404,
                    "Das angeforderte Mitglied konnte nicht gefunden werden!".to_string(),
                )
                .to_wrapper(),
            )),
            _ => Err(Json(
                ApiError::new(500, "Es ist ein Datenbankfehler aufgetreten!".to_string())
                    .to_wrapper(),
            )),
        },
    }
}

#[derive(Serialize)]
pub struct MemberListResult{
    members: Vec<MemberListEntryWithPermissions>,
    count: i64,
}

/// Read all members
/// Method: GET
/// URL-Parameters:
/// * limit: required, limit of search results
/// * offset: required
/// * groups: optional, comma seperated list of group uuid's. only members which are in at least one of these groups will be returned
/// * sort: optional, syntax: sort_field:[asc|desc]; example: &sort=firstname:asc
#[get("/api/members?<limit>&<offset>&<groups>&<sort>", format = "json")]
pub fn api_member_list(
    cookie: SessionCookie,
    settings: &State<Settings>,
    limit: i64,
    offset: i64,
    groups: Option<String>,
    sort: Option<String>,
) -> Result<Json<MemberListResult>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;
    if !caller.has_permission("modules.member_management.list.view".to_string()) {
        return Err(Json(
            ApiError::new(403, "Keine Berechtigung, Mitgliederliste abzurufen!".to_string())
                .to_wrapper(),
        ));
    };

    let mut mlr = MemberListResult{
        members: vec![],
        count: 0
    };

    let groups = match groups{
        Some(groups) => {
            debug!("filtering mm_list for groups: {}", groups);
            let groups = groups.split(",");

            let mut groups_res: Vec<uuid::Uuid> = vec![];

            for group in groups{
                if group != "" {
                    match uuid::Uuid::parse_str(group) {
                        Ok(uuidres) => {
                            groups_res.push(uuidres);
                        },
                        Err(e) => {
                            warn!("Couldn't parse group uuid in mm_list: {}", e);
                        }
                    }
                }
            }
            Some(groups_res)
        }
        None => None
    };

    let sort = match sort{
        Some(sort) => {
            match MemberListOrder::try_from(sort){
                Ok(sort) => Some(sort),
                Err(e) => {
                    warn!("Invalid sort request on member list: {}", e);
                    return Err(Json(ApiError::new(400, format!("Invalid sort argument: {}", e)).to_wrapper()))
                }
            }
        },
        None => None
    };

    match get_member_list(&settings, caller.entity_id, limit, offset, groups.clone(), sort) {
        Ok(result) => mlr.members = result,
        Err(e) => return Err(translate_diesel(e)),
    }
    match get_member_list_count(&settings, groups){
        Ok(c) => mlr.count = c,
        Err(e) => return Err(translate_diesel(e)),
    }

    Ok(Json(mlr))
}