use rocket::serde::json::Json;
use rocket::State;

use crate::database::controller::permissions::{get_role_permission_context, get_role_permission_id};
use crate::database::controller::roles::{get_members_for_role, get_permissions_for_role, get_roles};
use crate::database::model::api_members::RawMemberSearchResult;
use crate::database::model::permissions::Permission;
use crate::database::model::roles::Role;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::settings::Settings;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::parse_member_cookie;
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};

/// Get all permissions for role
///
/// # Api Call
/// * GET
/// * /api/roles/<role_id>/permissions
///
/// # Api Result
/// * Result JSON<Vec<[Permission]>> / ApiErrorWrapper
///
/// # Required permissions
/// * modules.settings.role_permissions.view
#[get("/api/roles/<role_id>/permissions", format = "json")]
pub fn get_role_permissions(
    settings: &State<Settings>,
    cookie: SessionCookie,
    role_id: String,
) -> Result<Json<Vec<Permission>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::settings::role_permissions::VIEW.to_string()){
        return Err(Json(ApiError::new(403, "No permission to read role permissions".to_string()).to_wrapper()))
    }

    match get_permissions_for_role(settings, role_id) {
        Ok(res) => Ok(Json(res)),
        Err(e) => Err(translate_diesel(e))
    }
}

/// Get context for permission role
///
/// # Api Call
/// * GET
/// * /api/roles/<role_id>/permissions/<permission>/context
///
/// # Api Result
/// * Result JSON<Vec<[uuid::Uuid]>> / ApiErrorWrapper
///
/// # Required permissions
/// * modules.settings.role_permissions.view
#[get("/api/roles/<role_id>/permissions/<permission>/context", format = "json")]
pub fn get_context_for_permission_role(
    settings: &State<Settings>,
    cookie: SessionCookie,
    permission: String,
    role_id: String,
) -> Result<Json<Vec<uuid::Uuid>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::settings::role_permissions::VIEW.to_string()){
        return Err(Json(ApiError::new(403, "No permission to read role permissions".to_string()).to_wrapper()))
    }

    let mut res : Vec<uuid::Uuid> = vec![];

    let rpi = get_role_permission_id(settings, &permission, &role_id);
        match rpi{
            Ok(rpi) => match rpi{
                Some(rpi) => {
                    match get_role_permission_context(settings, rpi){
                        Ok(context) => {
                            res.append(&mut context.clone())
                        },
                        Err(e) => return Err(translate_diesel(e))
                    }},
                None => {},
            },
            Err(e) => {
                return Err(translate_diesel(e))
            }
        }

    Ok(Json(res))
}


/// Get all roles
///
/// # Api Call
/// * GET
/// * /api/roles/
///
/// # Api Result
/// * Result JSON<Vec<[Role]>> / ApiErrorWrapper
///
/// # Required permissions
/// * modules.settings.role_permissions.view
#[get("/api/roles", format = "json")]
pub fn get_all_roles(
    settings: &State<Settings>,
    cookie: SessionCookie,
) -> Result<Json<Vec<Role>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::settings::role_permissions::VIEW.to_string()){
        return Err(Json(ApiError::new(403, "No permission to read roles".to_string()).to_wrapper()))
    }

    match get_roles(settings) {
        Ok(res) => Ok(Json(res)),
        Err(e) => Err(translate_diesel(e))
    }
}

/// Get all role members
///
/// # Api Call
/// * GET
/// * /api/roles/<role_id>/members
///
/// # Api Result
/// * Result JSON<Vec<[RawMemberSearchResult]>> / ApiErrorWrapper
///
/// # Required permissions
/// * modules.settings.role_permissions.view
#[get("/api/roles/<role_id>/members", format = "json")]
pub fn get_role_members(
    settings: &State<Settings>,
    cookie: SessionCookie,
    role_id: String,
) -> Result<Json<Vec<RawMemberSearchResult>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::settings::role_permissions::VIEW.to_string()){
        return Err(Json(ApiError::new(403, "No permission to read roles".to_string()).to_wrapper()))
    }

    match get_members_for_role(settings, &role_id) {
        Ok(res) => Ok(Json(res)),
        Err(e) => Err(translate_diesel(e))
    }
}

