use rocket::serde::json::Json;
use rocket::State;

use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::translate_diesel_error::translate_diesel;
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::modules::api::model::api_outcome::{ApiError, ApiErrorWrapper};
use crate::Settings;

#[derive(Serialize, Deserialize)]
pub struct PersonnelBilling {
    pub(crate) position_instance_id: uuid::Uuid,
    pub(crate) member_id: uuid::Uuid,
    pub(crate) fulfilled_time: i32,
    pub(crate) money_for_time: String,
    pub(crate) money_from_lump_sum: String,
    pub(crate) total_money: String,
}

#[get(
"/api/personnel_billing/<position_instance_id>",
format = "json"
)]
pub fn get_personnel_billing(
    settings: &State<Settings>,
    cookie: SessionCookie,
    position_instance_id: String,
) -> Result<Json<Option<PersonnelBilling>>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member.clone())?;
    //TODO: allow member_responsible for event to bypass this permission
    if !caller
        .has_permission(crate::permissions::modules::event_billing::personnel::VIEW.to_string())
    {
        return Err(Json(
            ApiError::new(
                403,
                "Keine Berechtigung Einsatzabrechnung anzusehen!".to_string(),
            )
                .to_wrapper(),
        ));
    }

    let pid = parse_uuid_string(position_instance_id)?;
    match crate::database::controller::billing::personnel_billing::read(settings, pid) {
        Ok(res) => {
            Ok(Json(match res {
                Some(res) => {
                    Some(PersonnelBilling {
                        position_instance_id: res.position_instance_id,
                        member_id: res.member_id,
                        fulfilled_time: res.fulfilled_time,
                        money_for_time: res.money_for_time.to_string(),
                        money_from_lump_sum: res.money_from_lump_sum.to_string(),
                        total_money: res.total_money.to_string(),
                    })
                },
                None => None
            }))
        },
        Err(e) => Err(translate_diesel(e))
    }
}