use rocket::State;
use crate::helper::settings::Settings;
use crate::helper::session_cookies::model::SessionCookie;
use rocket::serde::json::Json;
use crate::modules::api::model::api_outcome::{ApiErrorWrapper, ApiError};
use crate::database::model::vehicles::{VehicleCategory, Vehicle};
use crate::modules::api::member_management::controller::parser::{parse_member_cookie, parse_uuid_string};
use crate::database::controller::vehicles::{get_vehicle_categories, get_vehicles, get_vehicle_count, get_vehicle};
use crate::helper::translate_diesel_error::translate_diesel;
use crate::database::controller::members::check_access_to_resource;


#[get("/api/resources/vehicles/categories", format = "json")]
pub fn read_vehicle_categories(
    settings: &State<Settings>,
    cookie: SessionCookie,
) -> Result<Json<Vec<VehicleCategory>>, Json<ApiErrorWrapper>> {
    parse_member_cookie(cookie.member)?;
    match get_vehicle_categories(settings){
        Ok(category) => Ok(Json(category)),
        Err(e) => Err(translate_diesel(e))
    }
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct VehicleList{
    pub(crate) vehicle_list: Vec<Vehicle>,
    pub(crate) total_vehicle_count: i64,
}

#[get("/api/resources/vehicles?<entries>&<page>&<q>", format = "json")]
pub fn read_vehicle_list(
    settings: &State<Settings>,
    cookie: SessionCookie,
    entries: Option<i64>,
    page: Option<i64>,
    q: Option<String>
) -> Result<Json<VehicleList>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    if !caller.has_permission(crate::permissions::modules::resource_management::vehicles::list::VIEW.to_string()){
        return Err(Json(ApiError::new(403, "Keine Berechtigung Fahrzeugliste abzurufen!".to_string()).to_wrapper()))
    }

    //Apply default limit of 20 entries per page if none given
    let entries = match entries{
        Some(entries) => {
            if entries > 0{
                entries
            }else{
                20
            }
        }
        None => 20
    };

    //Apply default page 1 if no page given
    let page = match page{
        Some(page) => {
            if page > 0{
                page
            }else{
                1
            }
        }
        None => 1
    };
    
    let mut res = VehicleList{
        vehicle_list: vec![],
        total_vehicle_count: 0
    };

    match get_vehicles(settings, entries, (page-1)*entries, q){
        Ok(vehicles) => res.vehicle_list = vehicles,
        Err(e) => return Err(translate_diesel(e))
    };

    match get_vehicle_count(settings){
        Ok(count) => res.total_vehicle_count = count,
        Err(e) => return Err(translate_diesel(e))
    };

    Ok(Json(res))
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct ReducedVehicle{
    pub(crate) entity_id: uuid::Uuid,
    pub(crate) identifier: String,
    pub(crate) numberplate: Option<String>,
}

#[derive(Queryable, Clone, Deserialize, Serialize)]
pub struct ReadVehicleResponse{
    pub(crate) reduced_vehicle: Option<ReducedVehicle>,
    pub(crate) vehicle: Option<Vehicle>,
}

#[get("/api/resources/vehicles/<entity_id>", format = "json")]
pub fn read_vehicle(
    settings: &State<Settings>,
    cookie: SessionCookie,
    entity_id: String,
) -> Result<Json<ReadVehicleResponse>, Json<ApiErrorWrapper>> {
    let caller = parse_member_cookie(cookie.member)?;

    let vehicle = match get_vehicle(settings, parse_uuid_string(entity_id)?){
        Ok(vehicle) => vehicle,
        Err(e) => return Err(translate_diesel(e))
    };

    if check_access_to_resource(settings, caller.entity_id, vehicle.entity_id, crate::permissions::modules::resource_management::vehicles::core::VIEW) {
        Ok(Json(ReadVehicleResponse{
            reduced_vehicle: None,
            vehicle: Some(vehicle)
        }))
    }else{
        Ok(Json(ReadVehicleResponse{
            reduced_vehicle: Some(ReducedVehicle{
                entity_id: vehicle.entity_id,
                identifier: vehicle.identifier,
                numberplate: vehicle.numberplate
            }),
            vehicle: None,
        }))
    }
}