use rocket::http::Status;
use rocket::State;
use rocket_dyn_templates::Template;

use crate::database::controller::events::get_event;
use crate::helper::session_cookies::model::SessionCookie;
use crate::helper::sitebuilder::model::general::{Footer, Header, Script, Stylesheet};
use crate::helper::sitebuilder::model::sidebar::Sidebar;
use crate::modules::event_billing::event::EventBilling;
use crate::Settings;

/// Close Event UI
/// required permissions:
///     * modules.event_management.events.edit OR be member_responsible for event
#[get("/portal/eb/close_event?<event_id>")]
pub fn close_event(cookie: SessionCookie, settings: &State<Settings>, event_id: String) -> Result<Template, Status> {
    let caller = match cookie.member {
        //Unwraps member from cookie or send user to login if no member specified (user skipped member selection)
        Some(caller) => caller,
        None => return Err(Status::Unauthorized),
    };

    let event_id: uuid::Uuid = match event_id.parse() {
        Ok(event_id) => event_id,
        Err(e) => {
            warn!("Unparsable event_id in close_event: {}", e);
            return Err(Status::BadRequest);
        }
    };

    let event = match get_event(settings, event_id) {
        Ok(event) => event,
        Err(e) => {
            warn!("Couldn't load event: {}", e);
            return Err(Status::NotFound);
        }
    };

    if !caller.has_permission(crate::permissions::modules::event_management::events::EDIT.to_string()) {
        match event.member_responsible {
            Some(resp) => {
                if caller.entity_id != resp {
                    return Err(Status::Forbidden);
                }
            }
            None => {
                return Err(Status::Forbidden);
            }
        }
    }

    let header = Header {
        html_language: "de".to_string(),
        site_title: "Einsatzabrechnung".to_string(),
        stylesheets: vec![Stylesheet {
            path: "/css/errms.css".to_string(),
        }],
    };
    let footer = Footer {
        scripts: vec![Script {
            path: "/js/event_billing.js".to_string(),
        }],
    };
    let mut sidebar = Sidebar::new(caller.clone());
    sidebar.event_billing.active = true;

    Ok(Template::render("module_eb_close_event", EventBilling {
        header,
        footer,
        sidebar,
        caller: caller.entity_id,
        event,
    }))
}