use chrono::{Duration, Utc};
use rocket::http::{Cookie, CookieJar};
use rocket::State;

use crate::database::controller::login_protection::add_login_attempt;
use crate::database::controller::login_protection::login_attempts_exceeded;
use crate::database::controller::users::get_user_by_email;
use crate::database::model::users::User;
use crate::helper::session_cookies::model::SessionCookieStorage;
use crate::helper::settings::Settings;
use crate::modules::welcome::model::login_error_type::LoginError;
use crate::modules::welcome::model::login_form::LoginForm;

pub fn check_login(login_form: LoginForm, settings: &State<Settings>) -> Result<User, LoginError> {
    let user: User = match get_user_by_email(login_form.login_email.clone().to_lowercase(), &settings) {
        Some(user) => match login_attempts_exceeded(settings, login_form.login_email.to_lowercase()) {
            Ok(result) => {
                if result{
                    return Err(LoginError::MaxLoginAttemptsExceeded)
                }else{
                    user
                }
            },
            Err(_) => {
                return Err(LoginError::DatabaseError)
            }
        }
        None => {
            match login_attempts_exceeded(settings, login_form.login_email.to_lowercase()){
                Ok(result) => {
                    if result{
                        return Err(LoginError::MaxLoginAttemptsExceeded)
                    }else{
                        match add_login_attempt(settings, login_form.login_email.to_lowercase()){
                            Ok(_) => return Err(LoginError::UserNotFound),
                            Err(_) => return Err(LoginError::DatabaseError)
                        }
                    }
                },
                Err(_) => {
                    return Err(LoginError::DatabaseError)
                }
            }
        }
    };

    let password_hash = match user.password.clone(){
        None => {return Err(LoginError::NoPassword);}
        Some(pw) => pw
    };

    if argon2::verify_encoded(&password_hash, login_form.login_password.as_ref()).unwrap() {
        Ok(user)
    } else {
        match add_login_attempt(settings, login_form.login_email.to_lowercase()) {
            Ok(_) => return Err(LoginError::PasswordIncorrect),
            Err(_) => return Err(LoginError::DatabaseError)
        }
    }
}

pub fn add_session_cookie(
    user: User,
    settings: &State<Settings>,
    storage: &State<SessionCookieStorage>,
    cookies: &CookieJar,
) {
    let expires = Utc::now()
        .checked_add_signed(Duration::seconds(settings.application.session_timeout))
        .expect("Session timeout specified in configuration is too great!");
    let cookie = storage.add(expires, user, None);
    cookies.add_private(Cookie::new("session", cookie.id))
}
