use std::sync::Arc;

use lettre::Message;
use lettre::message::header;
use rocket::State;

use crate::database::controller::password_resets::add_token;
use crate::database::controller::users::get_user_by_email;
use crate::helper::mail_queue::queue::MailQueue;
use crate::helper::mail_templates::MailTemplates;
use crate::helper::settings::Settings;

#[derive(Serialize)]
pub struct PasswortResetMail {
    frontpage: String,
    email: String,
    reset_url: String,
}

/// Checks if email belongs to user, if so resets password
pub fn request_password_reset(settings: &State<Settings>, mt: &State<MailTemplates>, mq: &State<Arc<MailQueue>>, email: String) -> Result<(), ()>{
    let user = match get_user_by_email(email.clone(), settings){ //Check if email belongs to user, if not return
        Some(user) => user,
        None => return Ok(()),
    };

    let token = match add_token(settings, user.id){
        Ok(token) => token,
        Err(_) => return Err(()),
    };
    let pwrm = PasswortResetMail {
        frontpage: settings.application.url.clone(),
        email: email.clone(),
        reset_url: format!("{}password_reset?token={}", settings.application.url.clone(), token),
    };
    let body = match mt.registry.render("password-reset-de", &pwrm) {
        Ok(body) => body,
        Err(_) => return Err(()),
    };

    let email_address = match email.parse() {
        Ok(email) => email,
        Err(e) => {
            warn!("Couldn't parse password reset email: {}", e);
            return Err(())
        }
    };

    let mail = Message::builder()
        .from(settings.mail.from.clone().parse().unwrap())
        .reply_to(settings.mail.reply_to.clone().parse().unwrap())
        .to(email_address)
        .subject("Einsatz Online - Passwort Zurücksetzen") //TODO: use application name setting
        .header(header::ContentType::TEXT_PLAIN)
        .body(body).unwrap();

    match mq.add_mail(mail) {
        Ok(_) => Ok(()),
        Err(_) => Err(())
    }
}